/**
 *  Content:   InterferometerDataSourceJET
 *  Author:    Jakob Svensson (c) 2008 
 */

package aug.thomson;

import java.util.Arrays;

import base.SignalFetcher;
import jafama.FastMath;
import oneLiners.OneLiners;
import descriptors.JET.PPFDesc;
import descriptors.aug.AUGSignalDesc;
import mds.AugMDSFetcher;
import mds.JetMDSFetcher;
import algorithmrepository.Interpolation1D;
import algorithmrepository.LinearInterpolation1D;
import seed.minerva.deepend.Graph;
import seed.minerva.deepend.MinervaRuntimeException;
import seed.minerva.deepend.NodeImpl;
import seed.minerva.deepend.annotations.dependency;
import seed.minerva.deepend.annotations.dontSerialize;
import seed.minerva.diagnostics.ChannelDataEnable;
import seed.minerva.diagnostics.ServiceManager;
import seed.ws.xbase.server.XBase;
import seed.ws.xbase.types.Token;
import signals.JET.PPF;
import signals.aug.AUGSignal;

public class ThomsonVerticalAUG extends NodeImpl  {
	public static String DEFAULTNAMEPREFIX = "ThomsonVerticalAUG";
	public static final double neUncertainty = 1e18;
	public static final double TeUncertainty = 100;
	
	@dependency
	int pulse;
	
	@dependency
	double time;
	
	@dependency
	int[] enableRequest;
	
	@dependency
	boolean edgeSytem;

	// State
	// Pulse data
	double Z[];
	LinearInterpolation1D[] neInterp, TeInterp;
	LinearInterpolation1D[] neSigmaInterp, TeSigmaInterp;
	LinearInterpolation1D RInterp;
	int numChannels;
	
	// Timepoint data
	double R[];
	double ne[], neVar[];
	double Te[], TeVar[];
	
	ChannelDataEnable channelEnable;
	
				
	public ThomsonVerticalAUG() { this(null, null, null, null); }	
	public ThomsonVerticalAUG(Graph g, String name) { this(g, name, null, null); }	

	public ThomsonVerticalAUG(Graph g, String name, Object pulse, Object time) {
		super(g, name);
		
		if (pulse != null) set("pulse", pulse);
		if (time != null) set("time", time);
		set("edgeSytem", false);
		
	}
	
	@dontSerialize
	public void setEnableRequest(Object enableRequest) {
		set("enableRequest", enableRequest);		
	}
	
	public int[] getEnableRequest() {
		update("enableRequest");
		return enableRequest;
	}
	
	public void setEnableRequest(int enableRequestForAll) {
		update();
		setEnableRequest(OneLiners.fillIntArray(enableRequestForAll, channelEnable.getNumChannels()));		
	}
	
	public void setEnableRequest(int index, int request) {
		update();
		int[] newEnableRequest = null;
		if (!isSet("enableRequest") || enableRequest == null) {
			newEnableRequest = channelEnable.getEnableRequest();
		} else {
			newEnableRequest = enableRequest;
		}				
		
		newEnableRequest[index] = request;
		setEnableRequest(newEnableRequest);		
	}	
	
	public boolean[] getEnable() {
		update();
		return channelEnable.getEnable();
	}
			
	
	private void loadSignals() {
		SignalFetcher jmds = AugMDSFetcher.defaultInstance();
		
		String sysSelect = edgeSytem ? "_edge" : "_core";
		String shortSys = edgeSytem ? "_e" : "_c";
		
		AUGSignal neSig = (AUGSignal)jmds.getSig(new AUGSignalDesc(pulse, "VTA", "Ne" + shortSys));
		AUGSignal TeSig = (AUGSignal)jmds.getSig(new AUGSignalDesc(pulse, "VTA", "Te" + shortSys));
		AUGSignal neSigmaSig = (AUGSignal)jmds.getSig(new AUGSignalDesc(pulse, "VTA", "SigNe" + shortSys));
		AUGSignal TeSigmaSig = (AUGSignal)jmds.getSig(new AUGSignalDesc(pulse, "VTA", "SigTe" + shortSys));
		AUGSignal RSig = (AUGSignal)jmds.getSig(new AUGSignalDesc(pulse, "VTA", "R" + sysSelect));
		AUGSignal ZSig = (AUGSignal)jmds.getSig(new AUGSignalDesc(pulse, "VTA", "Z" + sysSelect));
				
		double t[] = (double[])neSig.getTVectorAsType(double.class); 
		
		double allNe[][] = (double[][])neSig.getDataAsType(double.class);
		double allTe[][] = (double[][])TeSig.getDataAsType(double.class);
		double allNeSigma[][] = (double[][])neSigmaSig.getDataAsType(double.class);
		double allTeSigma[][] = (double[][])TeSigmaSig.getDataAsType(double.class);
		double allR[] = (double[])RSig.getDataAsType(double.class);
		Z = (double[])ZSig.getDataAsType(double.class);
		
		numChannels = Z.length;
		if(channelEnable == null || channelEnable.getNumChannels() != numChannels)
			channelEnable = new ChannelDataEnable(numChannels);				
		
		RInterp = new LinearInterpolation1D(t, allR); //The R is smae for all channels, but moves a little over time
		
		neInterp = new LinearInterpolation1D[numChannels];
		TeInterp = new LinearInterpolation1D[numChannels];
		neSigmaInterp = new LinearInterpolation1D[numChannels];
		TeSigmaInterp = new LinearInterpolation1D[numChannels];
		int[] status = new int[numChannels];
		for(int i=0; i < numChannels; i++){
			neInterp[i] = new LinearInterpolation1D(t, allNe[i], Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
			TeInterp[i] = new LinearInterpolation1D(t, allTe[i], Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
			neSigmaInterp[i] = new LinearInterpolation1D(t, allNeSigma[i], Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
			TeSigmaInterp[i] = new LinearInterpolation1D(t, allTeSigma[i], Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
			status[i] = ChannelDataEnable.DATA_VALID;
		}
		
		channelEnable.setStatus(status);
		channelEnable.updateEnable();
	}
	
	
	private void updateSignals() {
		double neUnit = ServiceManager.getInstance().getUnitManager().getUnit("ElectronDensity");
		double TeUnit = ServiceManager.getInstance().getUnitManager().getUnit("ElectronTemperature");
		
		R = OneLiners.fillArray(RInterp.eval(time), numChannels);
		
		ne = new double[numChannels];
		neVar =  new double[numChannels];		
		Te = new double[numChannels];
		TeVar =  new double[numChannels];		
		for(int i=0;i<numChannels;++i) {
			ne[i] = neInterp[i].eval(time) / neUnit;
			Te[i] = TeInterp[i].eval(time) / TeUnit;
			neVar[i] = FastMath.pow2(neSigmaInterp[i].eval(time) / neUnit);
			TeVar[i] = FastMath.pow2(TeSigmaInterp[i].eval(time) / TeUnit);			
		}					
	}		
	

	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		//drop anything we can get or regenerate later
		RInterp = null; Z = null;
		neInterp = null; TeInterp = null;
		neSigmaInterp = null; TeSigmaInterp = null;
		ne = null; Te = null;
		neVar = null; TeVar = null;
		channelEnable = null;
		numChannels = -1;
	}
	
	public double[] getR() { update(); return R; }	
	public double[] getZ() { update(); return Z; }
	public double[][] getPosXYZ(){ update(); return new double[][]{ R, new double[numChannels], Z }; }
	
	public double[] getNeData() { update(); return ne; }
	public double[] getNeVariance() { update(); return neVar; }
	public double[] getTeData() { update(); return Te; }
	public double[] getTeVariance() { update(); return TeVar; }
	
	public int getNumChannels(){ update(); return numChannels; }	
	
	@Override
	public void updateState() {
		if (neInterp == null || isChanged("pulse" )) {
			loadSignals();
			ne = null;
		}
		if (isSet("enableRequest") && enableRequest != null) {
			channelEnable.setEnableRequest(enableRequest);
			channelEnable.updateEnable();
		}
		if (ne == null || isChanged("time")) {
			updateSignals();
		}
	}
	
}
