package seed.minerva.equilibrium.aug;

import base.SignalFetcher;
import binaryMatrixFile.BinaryMatrixFile;
import mds.AugMDSFetcher;
import descriptors.aug.AUGSignalDesc;
import oneLiners.OneLiners;
import algorithmrepository.Algorithms;
import algorithmrepository.LinearInterpolation1D;
import seed.function.ScalarFunction1D;
import seed.minerva.deepend.Graph;
import seed.minerva.deepend.NodeImpl;
import seed.minerva.deepend.annotations.dependency;
import seed.minerva.diagnostics.ServiceManager;
import signals.aug.AUGSignal;

/** Gives ff'(psi_N) for AUG equilibrium from the shot files */ 
public class AUGEquilibriumPpFFPSource extends NodeImpl {
	public static String DEFAULTNAMEPREFIX  = AUGEquilibriumPpFFPSource.class.getSimpleName();
	
	//uurrrg, need to deal with the signal's wierd time indexing
	private static final String equiFFPSignal= "FFP";
	private static final String equiPresSignal = "Pres";
	private static final String equiPsiInfoSignal = "PFxx";
	private static final String equiPsiSignal = "PFL";
	private static final String equiTimeSignal = "time";
	private static final String equiTimeIndexSignal = "ixti";
	
	//Parents
	@dependency private int pulse;
	@dependency private double time;
	@dependency private double psiMagAxis;
	@dependency private double psiLCFS;
	@dependency private String equiDiag;
	@dependency private String equiExp;
	
	private LinearInterpolation1D ffpInterp;
	private LinearInterpolation1D ppInterp;
	
	public final static double mu0 = 4.0*Math.PI*1.0e-7;
	public double ffpUnit = ServiceManager.getInstance().getUnitManager().getUnit("ffPrime");
	public double ppUnit = ServiceManager.getInstance().getUnitManager().getUnit("pPrime");
	
	private double pulseDataPsiInfo[][];
	private double pulseDataFFP[][];
	private double pulseDataPres[][];
	private double pulseDataPsi[][];
	private double pulseDataTimeOrdered[];
	private int indexInUnordered[];
	
	public AUGEquilibriumPpFFPSource() { this(null, null); }
	public AUGEquilibriumPpFFPSource(Graph g) { this(g, null); 	}
	public AUGEquilibriumPpFFPSource(Graph g, String name) { this(g, name, null, null); }	
	public AUGEquilibriumPpFFPSource(Graph g, String name, Object pulse, Object time) {
		super(g, name);
		
		if(pulse != null)
			set("pulse", pulse);
		
		if(time != null)
			set("time", time);
		
		set("equiDiag", "EQH");
	}
	
	
	private void loadPulseData() {
		
		SignalFetcher amds = AugMDSFetcher.defaultInstance();
		
		//AUGSignal IdxSig = (AUGSignal)amds.getSig(new AUGSignalDesc(pulse, equiDiag, equiTimeIndexSignal));
		//indexInUnordered = (float[])IdxSig.getData();
		
		AUGSignal TimeSig = (AUGSignal)amds.getSig(new AUGSignalDesc(pulse, equiDiag, equiTimeSignal));
		pulseDataTimeOrdered = (double[])TimeSig.getDataAsType(double.class);
		
		//reorder the time, in case it isnt
		indexInUnordered = new int[pulseDataTimeOrdered.length];
		Algorithms.quicksort(pulseDataTimeOrdered, indexInUnordered);
		
		AUGSignal PsiInfoSig = (AUGSignal)amds.getSig(new AUGSignalDesc(pulse, equiDiag, equiPsiInfoSignal, equiExp));
		pulseDataPsiInfo = (PsiInfoSig.getRank() == 3)  //deal with signal stupidity
							? ((double[][][])PsiInfoSig.getDataAsType(double.class))[0] 
							: (double[][])PsiInfoSig.getDataAsType(double.class);
		
		AUGSignal FFPSig = (AUGSignal)amds.getSig(new AUGSignalDesc(pulse, equiDiag, equiFFPSignal, equiExp));		
		pulseDataFFP = (FFPSig.getRank() == 3)  //deal with signal stupidity
							? ((double[][][])FFPSig.getDataAsType(double.class))[0] 
							: (double[][])FFPSig.getDataAsType(double.class);

		AUGSignal PresSig = (AUGSignal)amds.getSig(new AUGSignalDesc(pulse, equiDiag, equiPresSignal, equiExp));
		pulseDataPres = (PresSig.getRank() == 3)  //deal with signal stupidity
							? ((double[][][])PresSig.getDataAsType(double.class))[0] 
							: (double[][])PresSig.getDataAsType(double.class);

		AUGSignal PsiSig = (AUGSignal)amds.getSig(new AUGSignalDesc(pulse, equiDiag, equiPsiSignal, equiExp));
		pulseDataPsi = (PsiSig.getRank() == 3)  //deal with signal stupidity
							? ((double[][][])PsiSig.getDataAsType(double.class))[0] 
							: (double[][])PsiSig.getDataAsType(double.class);

		
	}
	
	
	private void selectTime() {
		int ti = OneLiners.getNearestIndex(pulseDataTimeOrdered, time);
		ti = indexInUnordered[ti]; //now an index to the rest of the signals
		
		//the equi code's flux values
		double equiPsiMaxAxis = pulseDataPsiInfo[ti][0];
		double equiPsiSeparatrix = pulseDataPsiInfo[ti][1];
		double equiPsiLimiter = pulseDataPsiInfo[ti][2];
		double equiPsi1 = (equiPsiSeparatrix > -1000) ? equiPsiSeparatrix : equiPsiLimiter;
		

		// ---- psi ----

		int n = pulseDataPsi[ti].length;
		double psiN[] = new double[n];
		for(int i=0; i < n; i++){
			psiN[i] = (pulseDataPsi[ti][i] - equiPsiMaxAxis) / (equiPsi1 - equiPsiMaxAxis);
		}
		
		int idx[] = new int[n];
		Algorithms.quicksort(psiN, idx);
				
		// ---- f f' ----		
		double ffp[] = new double[n];		
		
		for(int i=0; i < n; i++){
			ffp[i] = pulseDataFFP[ti][idx[i]];// / mu0 / ffpUnit;
			
			//change of units
			ffp[i] *= 2 * Math.PI * 1e6 / mu0;
			
			//this is dF/dPsi with psi as their psi which might be slightly different to ours
			//so we convert to dP/dPsiN first					
			ffp[i] *= (equiPsi1 - equiPsiMaxAxis) / (psiLCFS - psiMagAxis);
			
			ffp[i] /= ffpUnit;
		}
		
		ffpInterp = new LinearInterpolation1D(psiN, ffp);
		BinaryMatrixFile.mustWrite("/tmp/ffp", new double[][]{ psiN, ffp }, true);
		
		// ---- p' ----
		double pp[] = new double[n];
		
		for(int i=0; i < n; i++){
			pp[i] = pulseDataPres[ti][idx[i]*2+1]; // Pres is [P_0, P'_0,   P_1, P'_1  ... ]
			
			pp[i] *= 2 * Math.PI;
			
			//this is dP/dPsi with psi as their psi which might be slightly different to ours
			//so we convert to dP/dPsiN first
			pp[i] *= (equiPsi1 - equiPsiMaxAxis) / (psiLCFS - psiMagAxis);
			
			pp[i] /= ppUnit;
		}		
		
		ppInterp = new LinearInterpolation1D(psiN, pp);
		BinaryMatrixFile.mustWrite("/tmp/pp", new double[][]{ psiN, pp }, true);		
		
	}
	

	
	/*private void calcF(){
		int ti = OneLiners.getNearestIndex(pulseDataFFPTime, time);
		
		int n = pulseDataFFPPsi.length;
		
		double fPsiN[] = new double[n];
		double f[] = new double[n];
		
		// 2ff' is d( f^2 )/dPsi
		// integrating it w.r.t psi gives f^2
		// we integrate from the outside (psi = psiMax) with
		// the vacuum f as the constant
		f[n-1] = 0;
		fPsiN[n-1] = (pulseDataFFPPsi[n-1][ti] - psiMagAxis) / (psiLCFS - psiMagAxis);
		double fSqSum = f0*f0;
		for(int i=(n-1); i > 0; i++){
			//differential is w.r.t to actual psi 
			fSqSum += 2*pulseDataFFP[i][ti] * (pulseDataFFPPsi[i+1][ti] - pulseDataFFPPsi[i][ti]);
			f[i] = Math.sqrt(fSqSum);
			
			//but funtion is against psiN
			fPsiN[i] = (pulseDataFFPPsi[i][ti] - psiMagAxis) / (psiLCFS - psiMagAxis);
			
		}
		
		fInterp = new LinearInterpolation1D(fPsiN, f);
		
	}//*/
	
	@Override
	public void updateState() {

		if(pulseDataFFP == null || isChanged("pulse"))
			loadPulseData();
		
		if(ffpInterp == null || isChanged("time"))
			selectTime();
	}
	
	public double[] getFFPrime(double[] x) {
		update();
		
		return ffpInterp.eval(x);
	}

	public double[] getPPrime(double[] x) {
		update();
		
		return ppInterp.eval(x);
	}
}
