package seed.minerva.equilibrium.aug;

import java.lang.reflect.Array;

import base.SignalFetcher;
import algorithmrepository.Algorithms;
import algorithmrepository.CubicInterpolation2D;
import algorithmrepository.Interpolation1D;
import algorithmrepository.LinearInterpolation1D;

import oneLiners.OneLiners;
import descriptors.aug.AUGSignalDesc;
import mds.AugMDSFetcher;
import mds.AugShotfileFetcher;
import seed.minerva.diagnostics.ServiceManager;
import signals.aug.AUGSignal;
import sun.reflect.generics.reflectiveObjects.NotImplementedException;
import seed.minerva.deepend.Graph;
import seed.minerva.deepend.Node;
import seed.minerva.deepend.NodeImpl;
import seed.minerva.deepend.annotations.dependency;

/** Provides equilibrium for AUG from CLISTE run in the AUG shotfiles */
public class AUGEquilibriumSource extends NodeImpl {
	public static String DEFAULTNAMEPREFIX = AUGEquilibriumSource.class.getSimpleName();
	public final static double mu0 = 1.25663706e-6; // m kg s-2 A-2
	public static final double currentUnit = ServiceManager.getInstance().getUnitManager().getUnit("Current");
	
	
	@dependency private int pulse;
	@dependency private double time;
	
	/** The signal 'diagnostic' to read from (usually EQI, EQH or FPI?) */
	@dependency private String equiDiag;
	
	/** Renormalises total Ip to the correct value , only affects toroidalCurrentDensity() result */
	@dependency private boolean forceTotalIp;
	
	private String equiSignalPFMNode = "PFM";	
	//private String exp = "-pjm"; //P Mc.Carthy's
	@dependency private String equiExp;

	private String bvacSignalPath = "sig/MAI/BTF"; //toroidal magnetic field at R=1.65 m
	private double bvacMeasurementLocation = 1.65; // R / m

	private double pulseDataR[][];
	private double pulseDataZ[][];
	private double pulseDataPsi[][][];
	private double pulseDataIp[];
	/** Ordered time series */
	private double pulseDataTimeOrdered[];
	/** index into unordered time series of each ordered time point */
	private int indexInUnordered[];

	private double pulseDataBvacTime[], pulseDataBvac[];
	
	private int timeIndex = -1;
	private double Bvac0 = Double.NaN;
	private double psiGrid[][];
	private double gridR[], gridZ[];
	
	private CubicInterpolation2D interp;
	
	
	public AUGEquilibriumSource() { this(null, null, null, null); }	
	public AUGEquilibriumSource(String name) { this(null, name, null, null); }
	public AUGEquilibriumSource(Graph g, String name, Object pulse, Object time) {
		super(g, name);
		
		if(pulse != null)
			set("pulse", pulse);
		if(time != null)
			set("time", time);
		
		//set("equiDiag",  "EQI"); //normal equi
		set("equiDiag",  "EQH"); //slow high-res
		set("forceTotalIp", false);
		set("equiExp", "");
	}
	
	private void loadPulseData() {
		AUGSignalDesc desc;
		AUGSignal sig;
		
		//this can involve huge memory allocations, so drop everything else first
		tidyUpState();
		
		// The PFM signal is usually too big to read over MDS+
		//SignalFetcher amds = ServiceManager.getInstance().getAUGMDSFetcher();
		AugShotfileFetcher amds = AugShotfileFetcher.defaultInstance();
		
		desc = new AUGSignalDesc(pulse, equiDiag, "time", equiExp);
		sig = (AUGSignal)amds.getSig(desc);
		pulseDataTimeOrdered = (double [])sig.getDataAsType(double.class);
		
		try{
			desc = new AUGSignalDesc(pulse, equiDiag, "ixti", equiExp);
			sig = (AUGSignal)amds.getSig(desc);
			indexInUnordered = (int [])sig.getDataAsType(int.class);
			
		}catch(RuntimeException err){
			System.err.println("Couldn't read " + desc + ", because... ");
			err.printStackTrace();
			System.err.println("ordering time here...");
			
			//reorder the time series
			indexInUnordered = new int[pulseDataTimeOrdered.length];
			Algorithms.quicksort(pulseDataTimeOrdered, indexInUnordered);
			
		}
				
		
		desc = new AUGSignalDesc(pulse, equiDiag, "Ri", equiExp);
		sig = (AUGSignal)amds.getSig(desc);
		pulseDataR = ((double [][][])sig.getDataAsType(double.class))[0];
		
		desc = new AUGSignalDesc(pulse, equiDiag, "Zj", equiExp);
		sig = (AUGSignal)amds.getSig(desc);
		pulseDataZ = ((double [][][])sig.getDataAsType(double.class))[0];			
	
		try{
			desc = new AUGSignalDesc(pulse, equiDiag, "IpiPSI", equiExp);
			AUGSignal IpSig = (AUGSignal)amds.getSig(desc);
			Object o = IpSig.getDataAsType(double.class);
			while(!(o instanceof double[][])) ///yea, this is really annoying
				o = ((Object[])o)[0];
			o = OneLiners.transpose((double[][])o);
			pulseDataIp = ((double[][])o)[0];
			
		}catch(RuntimeException err){
			System.err.println("Couldn't read " + desc + ", because... ");
			err.printStackTrace();
			
			System.err.println("Interpolating from signal instead");
			
			desc = new AUGSignalDesc(pulse, "MAG", "Ipa");
			AUGSignal IpSig = (AUGSignal)amds.getSig(desc);
			double IpT[] = (double[])IpSig.getCoordsAsType(0, double.class);
			double IpV[] = (double[])IpSig.getDataAsType(double.class);
			LinearInterpolation1D ipInterp = new LinearInterpolation1D(IpT, IpV, Interpolation1D.EXTRAPOLATE_CONSTANT_VALUE, 0);
			pulseDataIp = ipInterp.eval(pulseDataTimeOrdered);
		}
		
		AUGSignal BVACSig = (AUGSignal)amds.getSig("/aug/" + pulse + "/" + bvacSignalPath);		
		pulseDataBvac = (double[])BVACSig.getDataAsType(double.class);
		pulseDataBvacTime = (double[])BVACSig.getTVectorAsType(double.class);
		
		//and finally, the big one - psiN
		desc = new AUGSignalDesc(pulse, equiDiag, equiSignalPFMNode, equiExp);
		System.out.print("Loading AUG Psi Map from signal '" + desc + "'... ");
		sig = (AUGSignal)amds.getSig(desc); //This can take a while, even from disk
		System.out.println("Done");
		pulseDataPsi = (double [][][])sig.getDataAsType(double.class);		
		
		
		psiGrid = null;
	
		
	}
	
	private void selectNearestTime() {
		
		if(time < pulseDataTimeOrdered[0] || time > pulseDataTimeOrdered[pulseDataTimeOrdered.length-1])
			throw new IllegalArgumentException("Time "+time+"out of range for equilibrium data. Data available for "+
					pulseDataTimeOrdered[0]+" < t < " + pulseDataTimeOrdered[pulseDataTimeOrdered.length-1] + "s");
		
		timeIndex = OneLiners.getNearestIndex(pulseDataTimeOrdered, time);
		System.out.println("AUGEquilibrium: Nearest time to req "+time+"s is index " + timeIndex + " at t="+pulseDataTimeOrdered[timeIndex]);

		
		int nT = pulseDataPsi.length;
		int nR = pulseDataR[0].length;
		int nZ = pulseDataZ[0].length;
		
		gridR = new double[nR];
		for(int i=0; i < nR; i++){
			gridR[i] = pulseDataR[indexInUnordered[timeIndex]][i];
		}
	
		gridZ = new double[nZ];
		for(int j=0; j < nZ; j++){
			gridZ[j] = pulseDataZ[indexInUnordered[timeIndex]][j];
		}
	
		psiGrid = new double[nR][nZ];
		for(int i=0; i < nR; i++){
			for(int j=0; j < nZ; j++){
				psiGrid[i][j] = pulseDataPsi[indexInUnordered[timeIndex]][j][i];
			}
		}
				
		int i = OneLiners.getNearestIndex(pulseDataBvacTime, time);
		Bvac0 = pulseDataBvac[i] * bvacMeasurementLocation;		
	}
		
	@Override
	public void updateState() {
		if(pulseDataPsi == null || isChanged("pulse") || isChanged("equiDiag"))
			loadPulseData();
		
		if(psiGrid == null || isChanged("time"))
			selectNearestTime();
		
	}
	
	@Override
	public void tidyUpState() {
		//super.tidyUpState(); NO AUTO!!!
		
		pulseDataR = null;
		pulseDataZ = null;
		pulseDataPsi = null;
		pulseDataIp = null;
		pulseDataTimeOrdered = null;
		indexInUnordered = null;
		pulseDataBvacTime = null;
		pulseDataBvac = null;
		psiGrid = null;
		gridR = null;
		gridZ = null;
		interp = null;
	}
	
	public double[][] getFluxGrid(){ update(); return psiGrid; }
	public double[] getGridR(){ update(); return gridR; }
	public double[] getGridZ(){ update(); return gridZ; }
	public double getToroidalVacuumFieldAt1m(){ update(); return Bvac0; }
	public double getEquiTime() { update(); return pulseDataTimeOrdered[timeIndex]; }
	public double[] getEquiTimebase() { update(); return pulseDataTimeOrdered; }
	
	
	/** Returns the plasma current as used by Cliste */
	public double getEquiPlasmaCurrent() { return pulseDataIp[timeIndex]; }
	
	// flux /= 2*pi
	
	public double[] toroidalCurrentDensity(double posXYZ[][]){		
		return toroidalCurrentDensityRPZ(OneLiners.XYZToRPhiZ(posXYZ));
	}
	
	public double[] toroidalCurrentDensityRPZ(double posRPZ[][]){
		update();
		
		if(interp == null)
			 interp = new CubicInterpolation2D(gridR, gridZ, psiGrid);
		
		if(isSet("forceTotalIp") && forceTotalIp)
			throw new NotImplementedException();
		
		double R[] = posRPZ[0];
		double Z[] = posRPZ[2];
				
		double j[] = new double[R.length];
		
		//calc the total plasma current in whatever these funny units are
		double totalCurrent = 0;
		
		double dR = 0.0001, dZ = 0.0001;
		double der[] = new double[3];
		double reNorm = 1.0 / (2 * Math.PI);
		
		for(int i=0; i < R.length; i++){			
			double psi = interp.eval(R[i], Z[i], der) * reNorm;
			
			//TODO: Shouldn't these be der[]??? 
			double psiRp = interp.eval(R[i] + dR, Z[i]) * reNorm;
			double psiRm = interp.eval(R[i] - dR, Z[i]) * reNorm;
			double psiZp = interp.eval(R[i], Z[i] + dZ) * reNorm;
			double psiZm = interp.eval(R[i], Z[i] - dZ) * reNorm;
			
			
			//equations are correct for SI units and psi = poloidal flux per unit in toroidal dir
			// J1 = (1/mu0) . d/dR (1/R . dPsi/dR)
			double J1 = -(1/mu0) * 	(
										(
												( (1/(R[i]+(dR/2))) * (psiRp - psi)/dR ) - 
												( (1/(R[i]-(dR/2))) * (psi - psiRm)/dR )    
										)
										/ dR
									);
			
			// J2 = 1/(mu0 R) . d^2Psi/dZ^2
			double J2 = -(1/(mu0 * R[i])) * (
												(
														(psiZp - psi)/dZ) - ((psi - psiZm)/dZ)
												) 
												/ dZ;
			
			j[i] = J1 + J2;
			
			j[i] /= currentUnit;
			
			//I[i] = J * w[i] * h[i];
			//if(!Double.isNaN(I[i]))
				//totalCurrent += I[i];		
		}
				
		/*System.out.println(" Done.\n" +
							"Total Ip from derivatives = " + totalCurrent + 
							(Double.isNaN(fixIp) ? "" : (", rescaling to " + fixIp) + "A"));
		*/
		
		
		//rescale to given total plasma current for equilibrium hacks
		//double currentScale = Double.isNaN(fixIp) ? 1.0 : (totalCurrent / fixIp);
		
		//put in our units, and clear NaNs (which usually means it is off grid, so set to 0)
		//for(int i=0; i < R.length; i++){			
		//	I[i] = currentScale * (Double.isNaN(I[i]) ? 0 : I[i]) / currentUnit;
		//}
	
		return j;
	}
	
}
