/**
 *  Content:   InterferometerDataSourceJET
 *  Author:    Jakob Svensson (c) 2008 
 */

package seed.minerva.interferrometry.aug;

import jafama.FastMath;
import oneLiners.OneLiners;
import descriptors.JET.PPFDesc;
import descriptors.aug.AUGSignalDesc;
import mds.AugMDSFetcher;
import mds.JetMDSFetcher;
import algorithmrepository.LinearInterpolation1D;
import seed.minerva.deepend.Graph;
import seed.minerva.deepend.MinervaRuntimeException;
import seed.minerva.deepend.NodeImpl;
import seed.minerva.deepend.annotations.dependency;
import seed.minerva.deepend.annotations.dontSerialize;
import seed.minerva.diagnostics.ChannelDataEnable;
import seed.minerva.diagnostics.ServiceManager;
import seed.ws.xbase.server.XBase;
import seed.ws.xbase.types.Token;
import signals.JET.PPF;
import signals.aug.AUGSignal;

public class InterferometerDataSourceAUG extends NodeImpl  {
	public static String DEFAULTNAMEPREFIX = "InterferometerDataSourceAUG";
	public static double maxStartEndDiff = 5e18;

	@dependency
	int pulse;
	
	@dependency
	double time;
	
	@dependency
	int[] enableRequest;
		
	// State
	double[] x1,y1,z1,x2,y2,z2;
	String[] losNames;
	LinearInterpolation1D[] signals;
	double[] data, var;	
	ChannelDataEnable channelEnable;
	int numChannels;

	/** Very hackey temporary fringe corrections */	
	double fringe = 1.143e19; //m^-2
	
	// Constants	
	public String diag = "DCN";
	public String signalNames[] = new String[]{  "V-1", "V-2", "V-3", "H-1", "H-2", "H-3", "H-4", "H-5" };
	
			
	public InterferometerDataSourceAUG() { this(null, null, null, null); }	
	public InterferometerDataSourceAUG(Graph g, String name) { this(g, name, null, null); }	

	public InterferometerDataSourceAUG(Graph g, String name, Object pulse, Object time) {
		super(g, name);
		
		if (pulse != null) set("pulse", pulse);
		if (time != null) set("time", time);
		
	}
	
	@dontSerialize
	public void setEnableRequest(Object enableRequest) {
		set("enableRequest", enableRequest);		
	}
	
	public int[] getEnableRequest() {
		update("enableRequest");
		return enableRequest;
	}
	
	public void setEnableRequest(int enableRequestForAll) {
		update();
		setEnableRequest(OneLiners.fillIntArray(enableRequestForAll, channelEnable.getNumChannels()));		
	}
	
	public void setEnableRequest(int index, int request) {
		update();
		int[] newEnableRequest = null;
		if (!isSet("enableRequest") || enableRequest == null) {
			newEnableRequest = channelEnable.getEnableRequest();
		} else {
			newEnableRequest = enableRequest;
		}				
		
		newEnableRequest[index] = request;
		setEnableRequest(newEnableRequest);		
	}	
	
	public boolean[] getEnable() {
		update();
		return channelEnable.getEnable();
	}
	
	private void loadGeom() {
		//from a picture in Alex Mlynek's AFS
		//V1 1.19 -1.3 1.19	1.3
		//V2 1.78 -1.3 1.78 1.3
		//V3 2.14 -1.3 2.14 1.3
		//H1: 2.6  0.16 1.0 0.14
		//H2: 2.6  0.32 1.0 0.3
		//H3: 2.6  -0.15 1.0 -0.13
		//H4: 2.7 -0.3 1.22 0.98
		//H5: 2.6  0.3 1.12 0.78
		
		losNames = signalNames.clone();
		
		x1 = new double[]{ 1.19, 1.78, 2.14, 	2.6, 2.6, 2.6, 2.7, 2.7 };
		y1 = new double[x1.length];
		z1 = new double[]{ -1.3, -1.3, -1.3, 	0.16, 0.32, -0.15, -0.3, 0.3 };
		
		x2 = new double[]{ 1.19, 1.78, 2.14, 	1.0, 1.0, 1.0, 1.22, 1.12 };
		y2 = new double[x1.length];
		z2 = new double[]{ 1.3, 1.3, 1.3, 		0.14, 0.3, -0.13, 0.98, 0.78 };
		numChannels = x1.length;
	}
		
	
	private void loadSignals() {
		AugMDSFetcher mds = AugMDSFetcher.defaultInstance();
		signals = new LinearInterpolation1D[numChannels];		
		int[] status = channelEnable.getStatus();
		for(int i=0;i<numChannels;++i) {
			AUGSignal s = null;
			signals[i] = null;			
			try {
				s = (AUGSignal)mds.getSig(new AUGSignalDesc(pulse, diag, signalNames[i]));
				//if (  ??? )
					//status[i] = ChannelDataEnable.DATA_NOT_VALID; 
					
				double[] d = (double[]) s.getDataAsType(double.class);
				double[] t = (double[]) s.getTVectorAsType(double.class);
						
				double offset = d[0];
				for(int j=0; j < d.length; j++)
					d[j] -= offset;
				
				signals[i] = new LinearInterpolation1D(t,d);
				
				if(FastMath.abs(d[d.length - 1]) > maxStartEndDiff){
					System.err.println("InterferometerDataSourceAUG: Invalidating channel '"+signalNames[i]+"' as it has a difference of "
									+ d[d.length - 1] + " ( > " + maxStartEndDiff + ")");
					
					status[i] = ChannelDataEnable.DATA_NOT_VALID;
				}
				
			} catch (RuntimeException ex) { 
				System.err.println("WARNING: "+ex.getMessage() + ", signal '"+diag+"/" + signalNames[i] + "' not avaliable.");
				status[i] = ChannelDataEnable.DATA_NOT_PRESENT; 				
			}	
		}	
		
		channelEnable.setStatus(status);
		channelEnable.updateEnable();
	}
	
	
	private void updateSignals() {
		double t = time;
		double neUnit = ServiceManager.getInstance().getUnitManager().getUnit("ElectronDensity");
		
		data = new double[numChannels];
		var =  new double[numChannels];		
		for(int i=0;i<numChannels;++i) {
			if (signals[i] == null) {
				data[i] = Double.NaN;
			} else {
				data[i] = signals[i].eval(t) / neUnit;
			}
			var[i] = Math.pow(1e17 / neUnit,2.0);			
		}					
	}		
	

	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		//drop anything we can get or regenerate later
		x1 = null; y1 = null; z1 = null;
		x2 = null; y2 = null; z2 = null;
		
		losNames = null; 
		signals = null;
		data = null; var = null;
		channelEnable = null;
		numChannels = 0;	
		
	}
	
	public double[] getLOS_x1() { update(); return x1; }	
	public double[] getLOS_y1() { update(); return y1; }	
	public double[] getLOS_z1() { update(); return z1; }	
	public double[] getLOS_x2() { update(); return x2; }	
	public double[] getLOS_y2() { update(); return y2; }	
	public double[] getLOS_z2() { update(); return z2; }	
	public String[] getLOS_names() { update(); return losNames; }	
	public double[] getData() { update(); return data; }
	public double[] getVariance() { update(); return var; }	
	public int getNumChannels(){ update(); return numChannels; }	

	private void autoChannelEnable() {
		int enReq[];
		switch(pulse){
			case 29388:
			default:
				enReq = new int[]{
						ChannelDataEnable.OFF,			//V-1
						ChannelDataEnable.OFF,			//V-2
						ChannelDataEnable.OFF,			//V-3
						
						ChannelDataEnable.AUTOMATIC,	//H-1
						ChannelDataEnable.AUTOMATIC,	//H-2
						ChannelDataEnable.AUTOMATIC,	//H-3
						ChannelDataEnable.AUTOMATIC,	//H-4
						ChannelDataEnable.AUTOMATIC,	//H-5
				};
		}
		
		
		/* Moved to data validity
		for(int i=0; i < numChannels; i++){
			if(enReq[i] != ChannelDataEnable.OFF){				
				double d[] = signals[i].getF();
				if(FastMath.abs(d[d.length - 1]) > maxStartEndDiff){
					System.err.println("InterferometerDataSourceAUG: Invalidating channel '"+signalNames[i]+"' as it has a difference of "
									+ d[d.length - 1] + " ( > " + maxStartEndDiff + ")");
					//enReq[i] = ChannelDataEnable.OFF;
					DATA_NOT_VALID
					
				}
			}
		}*/
		channelEnable.setEnableRequest(enReq);
		channelEnable.updateEnable();

	}
	
	@Override
	public void updateState() {
		if (signals == null || isChanged("pulse" )) {
			loadGeom();
			channelEnable = new ChannelDataEnable(numChannels);				
			loadSignals();
			autoChannelEnable();
			data = null;
		}
		if (isSet("enableRequest") && enableRequest != null) {
			channelEnable.setEnableRequest(enableRequest);
			channelEnable.updateEnable();
		}
		if (data == null || isChanged("time") || isChanged("fringeCorrections")) {
			updateSignals();
		}
	}

}
