package seed.minerva.magnetics.aug;

import javax.print.attribute.standard.PDLOverrideSupported;

import base.SignalFetcher;

import descriptors.JET.JPFDesc;
import mds.AugMDSFetcher;
import mds.JetMDSFetcher;
import algorithmrepository.Interpolation1D;
import algorithmrepository.LinearInterpolation1D;
import seed.minerva.deepend.Graph;
import seed.minerva.deepend.MinervaRuntimeException;
import seed.minerva.deepend.NodeImpl;
import seed.minerva.deepend.annotations.dependency;
import seed.minerva.diagnostics.ServiceManager;
import seed.minerva.diagnostics.UnitManager;
import signals.JET.JPF;
import signals.aug.AUGSignal;

/** Data source for PF (and TF) coils and PSL (Passive Stabilisation 'Line') */
public class AUGPFCoilCurrentsDataSource extends NodeImpl {
	public static String DEFAULTNAMEPREFIX = AUGPFCoilCurrentsDataSource.class.getSimpleName();

	private final static String pfModelIdXMLFile = "aug-pf.xml";
	private final static String pslModelIdXMLFile = "aug-psl.xml";
	
	private String pfModelId, pslModelId;
	
	/** Base path of signals for PF currents. /sig/MAI is the original and /sig/MBI seems to
	 * be processed a bit (offset subtracted at least) 
	 */
	//public final static String signalsPath = "/sig/MAI/";
	public final static String signalsPath = "/sig/MBI/";
		
	// Constants 
	/** PF Coil circuit - signal mapping
	 * Currently guessed 
	 */
	public final static String[][] pfCircuitInfo = new String[][] {
			{ "OH1",  signalsPath + "IOH", "1" },
			
			{ "OH2o", signalsPath + "IOH", "1", //Main OH current
					  signalsPath + "dIOH2s", "+1" }, //Difference from IOH to upper/oben
			{ "V1o", signalsPath + "IV1o", "1" },
			{ "CoAo", signalsPath + "ICoIo", "0" }, // Outer/Außer fast control
			{ "OH3o", signalsPath + "IOH", "1" },
			{ "CoOHo" },	//????
			{ "V2o", signalsPath + "IV2o", "1" },
			{ "V3o",  signalsPath + "IV3o", "1" },
			{ "CoIo",  signalsPath + "ICoIo", "1" }, //Inner fast control
			
			{ "OH2u", signalsPath + "IOH", "1",
				  	  signalsPath + "dIOH2s", "+1",	//Difference from IOH to upper/oben
					  signalsPath + "dIOH2u", "+1" },	 //Difference from upper/oben to lower/unten
			{ "V1u", signalsPath + "IV1u", "1" },
			{ "CoAu", signalsPath + "ICoIu", "0" }, //Outer/Außer fast control
			{ "OH3u", signalsPath + "IOH", "1" },
			{ "CoOHu" },	//????
			{ "V2u",  signalsPath + "IV2u", "1" },
			{ "V3u", signalsPath + "IV3u", "1" },
			{ "CoIu", signalsPath + "ICoIu", "1" }, //Inner fast control
			
			//Don't know what these are:
			//IV1oa	Signal	11 Vertical field Current 
			//IV1ua	Signal	12 Vertical field Current 

			//There is the signal /sig/MAI/ICoIu and I know that CoIu and CoAu are connected in parallel from one
			//power supply, so I've put the current in for both. Not sure if the current is in each, both or
			// if they have an equal share of the current.
			
			// CoOH is a very small coil next to OH3 and above V1. I can't find its signal.
	};
	

	/** This is the passive circuit, which goes around the top almost but not quite all
	 * the way, then down poloidally and back around the lower part and back up again.
	 * 
	 * I don't think there are the correct signals for this but it should probably
	 * be a free parameter here anyway.
	 */
	public final static String[][] pslCircuitInfo = new String[][] {
		{ "PSLu", "/sig/MAI/Ipslun", "-1" },
		{ "PSLo", "/sig/MAI/Ipslon", "1" },
	};
	
	//public static final String tfCurrentSignalName = "/sig/MAI/ITF";
	public static final String tfCurrentSignalName = "/sig/MAI/ITFsh"; //err.... what?
	private static final int numTFCoils = 16, numWindingsPerTFCoil = 24;
	
	LinearInterpolation1D totalTFCurrentSignal;
	double totalTFCurrent;
	
	// Parents
	@dependency private int pulse;
	@dependency private double time;
	
	
	LinearInterpolation1D[][] pfCurrentSignals;
	LinearInterpolation1D[] pslCurrentSignals;
	double[] pfCurrents, pslCurrents;
	
	double minTime, maxTime;
	
	public AUGPFCoilCurrentsDataSource() { this(null, null, null, null); }
	public AUGPFCoilCurrentsDataSource(Graph g, String name) { this(g, name, null, null); }
	public AUGPFCoilCurrentsDataSource(Graph g, String name, Object pulse, Object time) {
		super(g, name);
		
		if(pulse != null) 
			set("pulse", pulse);
		if(time != null) 
			set("time", time);
		
		pfModelId = "file://" + ServiceManager.getXBaseCachePath() + "/localXML/" + pfModelIdXMLFile;
		pslModelId = "file://" + ServiceManager.getXBaseCachePath() + "/localXML/" + pslModelIdXMLFile;
	}
	

	public String getPfModelId() { return pfModelId; }	
	public String getPslModelId() { return pslModelId; 	}
	
	public double[] getPfCurrents() { update(); return pfCurrents; }
	public double[] getPslCurrents() { update(); return pslCurrents;	}	
	public double getTotalTFCurrent(){ update(); return totalTFCurrent;	}
	
	private void loadCurrents() {
		minTime = Double.NEGATIVE_INFINITY;
		maxTime = Double.POSITIVE_INFINITY;
		try {
			SignalFetcher amds = AugMDSFetcher.defaultInstance();
			
			pfCurrentSignals = new LinearInterpolation1D[pfCircuitInfo.length][];
			for(int i=0;i<pfCurrentSignals.length;++i) {
				pfCurrentSignals[i] = new LinearInterpolation1D[(pfCircuitInfo[i].length-1)/2];
				
				for(int j=0; j < (pfCircuitInfo[i].length-1)/2; j++){
					if(pfCircuitInfo[i][1] != null) {
							
						AUGSignal sig = (AUGSignal)amds.getSig("aug/" + pulse + "/" + pfCircuitInfo[i][j*2+1]);
						double data[] = (double [])sig.getDataAsType(double.class);
						double time[] = (double [])sig.getTVectorAsType(double.class);
						
						double scale = Double.parseDouble(pfCircuitInfo[i][j*2+2]);
						if(scale != 1.0){
							for(int k=0; k < data.length; k++)
								data[k] *= scale;
						}
		
						pfCurrentSignals[i][j] = new LinearInterpolation1D(time, data, Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
						
						minTime = Math.max(minTime, time[0]);
						maxTime = Math.min(maxTime, time[time.length-1]);
					}
				}
			}
			
			pslCurrentSignals = new LinearInterpolation1D[pslCircuitInfo.length];
			for(int i=0;i<pslCurrentSignals.length;++i) {
				if(pslCircuitInfo[i][1] != null) {
						
					AUGSignal sig = (AUGSignal)amds.getSig("aug/" + pulse + "/" + pslCircuitInfo[i][1]);
					double data[] = (double [])sig.getDataAsType(double.class);
					double time[] = (double [])sig.getTVectorAsType(double.class);
	
					pslCurrentSignals[i] = new LinearInterpolation1D(time,data, Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
					
					minTime = Math.max(minTime, time[0]);
					maxTime = Math.min(maxTime, time[time.length-1]);
				}else{
					pslCurrentSignals[i] = null;
				}
				
			}
			
			AUGSignal tfSig = (AUGSignal)amds.getSig("aug/" + pulse + "/" + tfCurrentSignalName);
			double data[] = (double [])tfSig.getDataAsType(double.class);
			double time[] = (double [])tfSig.getTVectorAsType(double.class);
			
			double[] t = (double[]) tfSig.getTVectorAsType(double.class);
			double[] tfI = (double[]) tfSig.getDataAsType(double.class);
			double[] d = new double[t.length];
			for(int i=0;i<t.length;++i) {
				d[i] = tfI[i]*numTFCoils*numWindingsPerTFCoil;
			}
			
			totalTFCurrentSignal = new LinearInterpolation1D(t, d, Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
			
			minTime = Math.max(minTime, t[0]);
			maxTime = Math.min(maxTime, t[t.length-1]);
			
		} catch (Exception ex) {
			ex.printStackTrace();
			throw new MinervaRuntimeException("Error reading AUG PF coil current signals: "+ex.getMessage());
		}
		
		pslCurrents = null;
		pfCurrents = null;
	}
	
	private void updateCurrents(double time) {
		double currentUnit = ServiceManager.getInstance().getUnitManager().getUnit("Current");
		pfCurrents = new double[pfCurrentSignals.length];
		for(int i=0;i<pfCurrents.length;++i) {
			for(int j=0; j < pfCurrentSignals[i].length; j++){
				pfCurrents[i] += pfCurrentSignals[i][j].eval(time)*Double.parseDouble(pfCircuitInfo[i][j*2+2])/currentUnit;
			}
		}		
		
		pslCurrents = new double[pslCurrentSignals.length];
		for(int i=0;i<pslCurrents.length;++i) {
			if(pslCurrentSignals[i] != null)
				pslCurrents[i] = pslCurrentSignals[i].eval(time)*Double.parseDouble(pslCircuitInfo[i][2])/currentUnit;
			else
				pslCurrents[i] = 0;
		}
		
		if (totalTFCurrentSignal == null) {
			totalTFCurrent = Double.NaN;
		} else {
			totalTFCurrent = totalTFCurrentSignal.eval(time)/currentUnit;
		}	
	}

	@Override
	public void updateState() {
		
		if(pfCurrentSignals == null || isChanged("pulse")){ //also fire when we tidied these up
			loadCurrents();
		} 
		
		if (pfCurrents == null || isChanged("time")) {
			updateCurrents(time);
		}
	}
		
	@Override
	public void tidyUpState() {
		super.tidyUpState();
				
		//drop things that are big and we can regenerate
		totalTFCurrentSignal = null;
		pslCurrentSignals = null;
		pfCurrentSignals = null;
		pfCurrents = null;
	}
	
	public double getMinTime(){ update(); return minTime; }
	public double getMaxTime(){ update(); return maxTime; }	
	
}
