package seed.minerva.magnetics.aug;

import jafama.FastMath;

import java.io.FileInputStream;


import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;

import binaryMatrixFile.BinaryMatrixFile;

import algorithmrepository.Algorithms;

import oneLiners.OneLiners;

import seed.magnetostatics.types.InfiniteStraightCurrent;
import seed.magnetostatics.mdd.CompilerForObjectModel;
import seed.magnetostatics.mdd.MagnetostaticsObjectModel;
import seed.magnetostatics.mdd.RuntimeJava;
import seed.magnetostatics.mdd.Util;
import seed.magnetostatics.types.Coil;
import seed.magnetostatics.types.MagneticConfiguration;
import seed.magnetostatics.types.PolygonFilament;
import seed.magnetostatics.types.SerialCircuit;
import seed.magnetostatics.types.SolidCircularArc;
import seed.matrix.DenseMatrix;
import seed.matrix.EigenvalueDecomposition;
import seed.minerva.MinervaSettings;
import seed.minerva.diagnostics.ServiceManager;
import seed.minerva.magnetics.MagnetostaticsResponsesLayer;
import seed.ws.math.types.Points3D;
import seed.ws.math.types.Vector;

/** Makes the XML model for the AUG PF coils */
public class GenerateAUGPFCoilModel {
	
	/** This is taken for the moment from the "1990 coil measurements" in the AUG ringbuch:
	 * https://www.aug.ipp.mpg.de/documentation/pub/MagnetSystem/PoloidalFieldSystem/5MeasuredCoilPositions1990.pdf
	 * 
	 * There are two sets, one for 'isolierten Spulen' and one for 'blanken Spulenwickel'.
	 * I think the second one translates as 'bare windings', so I'm using those, which are:
	 * 
	 * The letters on the end are 'o' for oben (above) and 'u' for unten (below)
	 * 
	 * The number of windings information is probably wrong.
	 * 
	 * There are a few different sources on the AUG website and they all disagree, by quite a lot.
	 * 
	 Upper Coils:
        Name    Left    Bot     Right   Top     Width   Height		Turns		Desc
		OH1       252     -1510   502     1510    250     3020		510		Ohmic 1 
		
		Upper/Oben
		OH2o      492     1800    830     2214.5  338     414.5		81		Ohmic 2 Upper
		V1o(a)    1274.5  2185.0  1697.2  2531.1  422.7   346.1		66		Vert. Field 1 Upper part a
		V1o(b)    1871    2185.0  2043.4  2531.3  172.4   346.3		27		Vert. Field 1 Upper part b
		CoAo      2637.8  1783.8  2775.7  1988.8  137.9   205		23		Outer/Außer fast control coil
		OH3o      2939    1981.7  3084.8  2034.5  145.8   52.8		5		Ohmic 3 Upper
		CoOHo     3153.8  1981.7  3329.2  2034.5  175.4   52.8		6		? Ohmic Upper
		V2o(a)    2931.5  1466    3090.5  1897    159     431		39		Vert. Field 2 Upper part a
		V2o(b)    3149    1466    3343.5  1897    194.5   431		47		Vert. Field 2 Upper part b
		V3o       3475.5  601.9   3723.4  785.5   247.9   183.6		28		Vert. Field 3 Upper
		CoIo(a)   2435.5  780     2510.5  812     75      32		1		 			
		CoIo(b)   2454.2  740.5   2529.2  772.5   75      32		1		Inner fast control coil
		CoIo(c)   2473.0  701     2548.0  733     75      32		1			Little stepped coil Upper parts a-e
		CoIo(d)   2491.8  661.5   2566.8  693.5   75      32		1
		CoIo(e)   2510.5  622     2585.5  654     75      32		1
		
		Lower/Unten (although the coords are as if they were upper)
		OH2u      492     1800    830     2214.5  338     414.5		81		Ohmic 2 Upper
		V1u(a)    1272.7  2185    1694.6  2531.6  421.9   346.6		66		Vert. Field 1 Upper part a
		V1u(b)    1870.2  2185    2042.2  2531.6  172     346.6		27		Vert. Field 1 Upper part b
		CoAu      2637.8  21784.2 2775.5  1989.2  137.7   205		23		Outer/Außer fast control coil
		OH3u      2938.3  1982.2  3084.1  2035    145.8   52.8		5		Ohmic 3 Upper
		CoOHu     3156.4  1982.2  3331.8  2035    175.4   52.8		6		? Ohmic Upper
		V2u(a)    2932.5  1466    3091.5  1897    159     431		39		Vert. Field 2 Upper part a
		V2u(b)    3148.5  1466    3343    1897    194.5   431		47		Vert. Field 2 Upper part b
		V3u       3476.1  599.6   3724.1  783.2   248     183.6		28		Vert. Field 3 Upper
		CoIu(a)   2433.8  780.    2508.8  812.    75.     32.		1		
		CoIu(b)   2452.5  740.5   2527.5  772.5   75.     32.		1		Inner fast control coil
		CoIu(c)   2471.3  701.    2546.3  733.    75.     32.		1			Little stepped coil Upper parts a-e 
		CoIu(d)   2490.1  661.5   2565.1  693.5   75.     32.		1
		CoIu(e)   2508.8  622.    2583.8  654.    75.     32.		1
*/
	
	public final static String names[] = new String[]{
			"OH1", 
			"OH2o", "V1o", "CoAo", "OH3o", "CoOHo", "V2o", "V3o",  "CoIo", 
			"OH2u", "V1u", "CoAu", "OH3u", "CoOHu", "V2u", "V3u", "CoIu", 
	};
	
	public final static double coords[][][] = new double[][][] {
			{{ 252, -1510, 502, 1510, 250, 3020,	510 }}, 			// OH1
			
			{{ 492, 1800, 830, 2214.5, 338, 414.5, 81 }},				// OH2o
			{
				{ 1274.5, 2185, 1697.2, 2531.1, 422.7, 346.1, 66 },		// V1o(a)
				{ 1871, 2185, 2043.4, 2531.3, 172.4, 346.3, 27 }		// V1o(b)
			},
			{{ 2637.8, 1783.8, 2775.7, 1988.8, 137.9, 205, 23 }},		// CoAo
			{{ 2939, 1981.7, 3084.8, 2034.5, 145.8, 52.8, 5 }},			// OH3o
			{{ 3153.8, 1981.7, 3329.2, 2034.5, 175.4, 52.8, 6 }},		// CoOHo
			{
				{ 2931.5, 1466, 3090.5, 1897, 159, 431, 39 },			//V2o(a)
				{ 3149, 1466, 3343.5, 1897, 194.5, 431, 47 },			//V2o(b)
			},
			{{ 3475.5, 601.9, 3723.4, 785.5, 247.9, 183.6, 28 }},		//V3o	
			{
				{ 2435.5, 780, 2510.5, 812, 75, 32, 1 },				//CoIo(a)
				{ 2454.2, 740.5, 2529.2, 772.5, 75, 32, 1 },			//CoIo(b)
				{ 2473, 701, 2548, 733, 75, 32, 1 },					//CoIo(c)
				{ 2491.8, 661.5, 2566.8, 693.5, 75, 32, 1 },			//CoIo(d)
				{ 2510.5, -654, 2585.5, -622, 75, 32, 1 },				//CoIo(e)
			},
			
			{{ 492, -2214.5, 830, -1800, 338, 414.5, 81 }},				// OH2u
			{
				{ 1272.7, -2531.6, 1694.6, -2185, 421.9, 346.6, 66 },	// V1u(a)
				{ 1870.2, -2531.6, 2042.2, -2185, 172, 346.6, 27 }		// V1u(b)
			},
			{{ 2637.8, -1989.2, 2775.5, -1784.2, 137.7, 205, 23 }},		// CoAu
			{{ 2938.3, -2035, 3084.1, -1982.2, 145.8, 52.8, 5 }},		// OH3u
			{{ 3156.4, -2035, 3331.8, -1982.2, 175.4, 52.8, 6 }},		// CoOHu
			{
				{ 2932.5, -1897, 3091.5, -1466, 159, 431, 39 },			// V2u(a)
				{ 3148.5, -1897, 3343, -1466, 194.5, 431, 47 },			// V2u(b)
			},
			{{ 3476.1, -783.2, 3724.1, -599.6, 248, 183.6, 28 }},		// V3u
			{
				{ 2433.8, -812, 2508.8, -780, 75, 32, 1 },				// CoIu(a)
				{ 2452.5, -772.5, 2527.5, -740.5, 75, 32, 1 },			// CoIu(b)
				{ 2471.3, -733, 2546.3, -701, 75, 32, 1 },				// CoIu(c)
				{ 2490.1, -693.5, 2565.1, -661.5, 75, 32, 1 },			// CoIu(d)
				{ 2508.8, -654, 2583.8, -622, 75, 32, 1 },				// CoIu(e)
			}
		};
	
	public static void main(String[] args) {
		MinervaSettings.instance();
		
		//processCoilDrawing();
		
		//tfFieldTest();
		//if(true)return;
		
		String xmlPath = ServiceManager.getXBaseCachePath() + "/localXML";
		MagneticConfiguration pfs = makePFModel();
		writeXML(xmlPath + "/aug-pf.xml", pfs);
		
		MagneticConfiguration psl = makePSLModel();
		writeXML(xmlPath + "/aug-psl.xml", psl);
		
	}
	
	public static void writeXML(String fileName, MagneticConfiguration conf){

		OneLiners.makePath(fileName);
		PrintWriter out;
		try {
			out = new PrintWriter(fileName);
			out.print(seed.magnetostatics.mdd.Util.toXML(conf));
			out.close();
			
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	
	public static final double pslCoords[][] = new double[][]{
		// Zbot, Ztop, Rbotleft, Rbotright, Rtopleft, Rtopright, 
		{ -0.87, -0.56, 1.88, 1.97, 2.08, 2.17 }, // top / oben 
		{ 0.56, 0.87, 2.15, 2.24, 1.96, 2.05 },  // bottom / unten
	};
	static String pslNames[] = { "PSLu", "PSLo" };
	
	/** Passive Stabilisation Conductor/Leitung */
	public static MagneticConfiguration makePSLModel(){

		int numSegments = 10;
		
		MagneticConfiguration configuration = new MagneticConfiguration();
		configuration.setName("aug.circuits.psl");			
		
		for(int i=0; i < pslCoords.length; i++) {
			SerialCircuit circuit = new SerialCircuit();
			circuit.setName(pslNames[i]);							
			configuration.getCircuit().add(circuit);
			
			double Z_b = pslCoords[i][0];
			double Z_t = pslCoords[i][1];
			double R_bl = pslCoords[i][2];
			double R_br = pslCoords[i][3];
			double R_tl = pslCoords[i][4];
			double R_tr = pslCoords[i][5];
			            	
			Coil coil = new Coil();
			coil.setName(pslNames[i]);
			coil.setNumWindings(1);
			coil.setWindingDirection(1);
			
			for(int j=0; j < numSegments; j++){
				
				double f0 = j * (1.0 / numSegments);
				double f1 = (j+1) * (1.0 / numSegments);
				
				double segZ_b = Z_b + f0 * (Z_t - Z_b);
				double segZ_t = Z_b + f1 * (Z_t - Z_b);
				
				double segR_bl = R_bl + f0 * (R_tl - R_bl);
				double segR_tl = R_bl + f1 * (R_tl - R_bl);
				
				double segR_br = R_br + f0 * (R_tr - R_br);
				double segR_tr = R_br + f1 * (R_tr - R_br);
				
				double segR_l = (segR_bl + segR_tl) / 2.0;
				double segR_r = (segR_br + segR_tr) / 2.0;
				
				double W = segR_r - segR_l;
				double H = segZ_t - segZ_b;
				
				double R0 = (segR_r + segR_l) / 2;
				double Z0 = (segZ_b + segZ_t) / 2;
								
				SolidCircularArc beam = new SolidCircularArc();
				beam.setName(names[i]);
				beam.setArcangle(2*Math.PI);
				beam.setNormal(vector(0, 0, 1));
				beam.setR0(vector(R0, 0, Z0));
				beam.setCentre(vector(0, 0, Z0));
				beam.setHeight(H);
				beam.setWidth(W);
				coil.getCurrentCarrierPrimitive().add(beam);
				//coil.getCurrentCarrierRatios().add(1.0);
			}
			circuit.getCurrentCarrier().add(coil);
		}
		
		return configuration;
	}
		
	public static MagneticConfiguration makePFModel(){
	
		MagneticConfiguration configuration = new MagneticConfiguration();
		configuration.setName("aug.circuits.pf");			
		
		int numCircuits = coords.length;
				
		for(int i=0;i<numCircuits;++i) {
			SerialCircuit circuit = new SerialCircuit();
			circuit.setName(names[i]);										
			configuration.getCircuit().add(circuit);
			
			int numCoils = coords[i].length;
			
			for(int j=0; j < numCoils; j++){
			
				double R0 = (coords[i][j][2] + coords[i][j][0]) / 2;
				double Z0 = (coords[i][j][3] + coords[i][j][1]) / 2;
				double W = coords[i][j][4];
				double H = coords[i][j][5];
				int nWindings = (int)coords[i][j][6];
				
				assert(W == (coords[i][j][2] - coords[i][j][0]));
				assert(H == (coords[i][j][3] - coords[i][j][1]));
				
				//they were in mm
				R0 /= 1000;
				Z0 /= 1000;
				W /= 1000;
				H /= 1000;
				
				Coil coil = new Coil();
				coil.setName(names[i]);
				coil.setNumWindings(nWindings);
				coil.setWindingDirection(1);
				
				SolidCircularArc beam = new SolidCircularArc();
				beam.setName(names[i]);
				beam.setArcangle(2*Math.PI);
				beam.setNormal(vector(0, 0, 1));
				beam.setR0(vector(R0, 0, Z0));
				beam.setCentre(vector(0, 0, Z0));
				beam.setHeight(H);
				beam.setWidth(W);
				coil.getCurrentCarrierPrimitive().add(beam);
				circuit.getCurrentCarrier().add(coil);
			}
		}
		
		return configuration;
	}
	
	public static Vector vector(double x, double y, double z) {
		return CompilerForObjectModel.vector(new double[] { x, y, z });
	}
	
	public static void processCoilDrawing(){
		String fileName = "/work/ipp/scrap/coils/all-40-32i-32a.txt";
		
		String lines[] = OneLiners.linesFromFile(fileName);
		
		ArrayList<double[][]> data[] = new ArrayList[4];
		
		for(int j=0; j < 4; j++)
			data[j] = new ArrayList<double[][]>();
		
			
		PrintWriter pw;
		try {
			pw = new PrintWriter("/work/ipp/scrap/coils/all.txt");
		} catch (FileNotFoundException e) {
			throw new RuntimeException(e);
		}
		
		for(int i=0; i < (lines.length/2); i++){
			int type = Integer.parseInt(lines[i*2+1]);
			
			String dStr[] = lines[i*2].split(",");
			double x[] = new double[dStr.length/2];
			double y[] = new double[dStr.length/2];
			for(int j=0; j < x.length; j++) {
				x[j] = Double.parseDouble(dStr[2*j]);
				y[j] = Double.parseDouble(dStr[2*j+1]);
				if(j>0){
					x[j] += x[j-1];
					y[j] += y[j-1];
				}
				
			}
			System.out.println(type + "\t" + x.length);
			
			data[type].add(new double[][]{ x, y });
			
			//find the mean X and Y for centre of coil
			double meanX = 0, meanY = 0;
			int nOK=0;
			boolean ok[] = new boolean[x.length];
			double minLen2 = 0.01*0.01;
			for(int j=0; j < x.length; j++){
				ok[j] = true;
				for(int k=0; k < j; k++){
					double len2 = (x[j]-x[k])*(x[j]-x[k])+(y[j]-y[k])*(y[j]-y[k]);
					if(len2 < minLen2){
						ok[j] = false;
						break;
					}
				}
				if(ok[j]){
					meanX += x[j];
					meanY += y[j];
					nOK++;
				}
			}
			meanX /= nOK;
			meanY /= nOK;
			
			//for each point, find the mean angle
			/*double meanAng = 0;
			double totalLen = 0;
			for(int j=1; j < (x.length-1); j++){
				double ang = Math.atan2(y[j] - meanY, x[j] - meanX) ;
				double len = (x[j]-meanX)*(x[j]-meanX) + (y[j]-meanY)*(y[j]-meanY);
				//len *= len;
				meanAng += len * ((ang < 0) ? (ang + Math.PI) : ang);
				totalLen += len;
			}
			meanAng /= totalLen;
			
			//pw.println(type +"\t"+ x[0] +"\t"+ y[0] +"\t"+ Math.cos(meanAng) +"\t" +  Math.sin(meanAng));
			//*/
			
			//calc the covariance, from which we'll find the principal dir
			double c[] = new double[3];
			for(int j=0; j < (x.length-1); j++){
				if(ok[j]){
					c[0] = (x[j] - meanX) * (x[j] - meanX);
					c[1] = (x[j] - meanX) * (y[j] - meanY);
					c[2] = (y[j] - meanY) * (y[j] - meanY);
				}
			}
			c[0] /= nOK;
			c[1] /= nOK;
			c[2] /= nOK;
			
			DenseMatrix m = new DenseMatrix(new double[][]{{ c[0], c[1] }, { c[1], c[2] }});
			EigenvalueDecomposition e = new EigenvalueDecomposition(m);
			
			double ve[][] = e.getEigenvectors().toArray();
			double va[] = e.getRealEigenvalues();
			double vec[] = (va[0] > va[1]) ? ve[0] : ve[1];
			
			double nx = (x[2]-x[1]);
			double ny = (y[2]-y[1]);			
			double nl = Math.sqrt(nx*nx+ny*ny);
			nx /= nl; ny /= nl;
			
			double cx = meanX - 360, cy = meanY - 705;
			double cl = Math.sqrt(cx*cx+cy*cy);
			cx /= cl; cy /= cl;
			
			double ncAng = Math.acos(nx*cx + ny*cy);
			if(ncAng > (Math.PI/4)){
				nl = nx;
				nx = -ny;
				ny = nl;
			}//*/
			
			double ang = Math.atan2(ny, nx);
			
			pw.println(type +"\t"+ meanX +"\t"+ meanY +"\t"+ nx +"\t" +  ny + "\t" + ang);
			//*/
		}
		pw.close();
	}

	
	/** This is not right for AUG, just circles */
	public static MagneticConfiguration simpleTFModel(){

		MagneticConfiguration configuration = new MagneticConfiguration();
		configuration.setName("aug.circuits.circTF");			
		
		SerialCircuit circuit = new SerialCircuit();
		circuit.setName("tf");										
		configuration.getCircuit().add(circuit);
		
		int nCoils = 16;
		int nWindingsPerCoil = 24;
		
		double rOuter[] = new double[]{ 0.52,  0.52,  0.55,  0.6 ,  0.7 ,  0.9 ,  1.1 ,  1.25,  1.55,  1.8 ,  2.  ,  2.2 ,  2.35,  2.6 ,  2.7 ,  2.85,  2.95,  3.05,  3.15,  3.2 ,  3.22 };
		double zOuter[] = new double[]{ 0.1  ,  1.  ,  1.25,  1.4 ,  1.55,  1.75,  1.85,  1.9 ,  1.92,  1.9 ,  1.85,  1.75,  1.65,  1.45,  1.35,  1.15,  1.  ,  0.8 ,  0.5 ,  0.25,  0.1 };
		double rInner[] = new double[]{ 0.9 ,  0.9 ,  0.9 ,  0.95,  1.  ,  1.1 ,  1.2 ,  1.5 ,  1.9 ,  2.05,  2.2 ,  2.45,  2.6 ,  2.65,  2.75,  2.8 ,  2.8 };
		double zInner[] = new double[]{ 0.1  ,  1.  ,  1.1 ,  1.2 ,  1.3 ,  1.4 ,  1.45,  1.5 ,  1.45,  1.35,  1.25,  1.  ,  0.8 ,  0.7 ,  0.45,  0.2 ,  0.1 };
		double rMid[] = new double[]{ 0.73,0.73,0.8,0.9,1.15,1.35,1.75,1.9,2.15,2.35,2.6,2.75,2.85,2.95,3,3.03 };
		double zMid[] = new double[]{ 0.1, 1.15,1.35,1.5,1.65,1.7,1.7,1.65,1.55,1.4,1.15,0.9,0.75,0.5,0.3,0.1 };
		
		double rOuterFull[] = Arrays.copyOf(rOuter, rOuter.length*2);
		double zOuterFull[] = Arrays.copyOf(zOuter, zOuter.length*2);
		for(int k=0; k < rOuter.length; k++){
			rOuterFull[k] = rOuter[k];
			rOuterFull[rOuter.length*2 - k - 1] = rOuter[k];
			zOuterFull[k] = zOuter[k];
			zOuterFull[zOuter.length*2 - k - 1] = -zOuter[k];
		}
		
		double rInnerFull[] = Arrays.copyOf(rInner, rInner.length*2);
		double zInnerFull[] = Arrays.copyOf(zInner, zInner.length*2);
		for(int k=0; k < rInner.length; k++){
			rInnerFull[k] = rInner[k];
			rInnerFull[rInner.length*2 - k - 1] = rInner[k];
			zInnerFull[k] = zInner[k];
			zInnerFull[zInner.length*2 - k - 1] = -zInner[k];
		}
		
		double rMidFull[] = Arrays.copyOf(rMid, rMid.length*2+1);
		double zMidFull[] = Arrays.copyOf(zMid, zMid.length*2+1);
		for(int k=0; k < rMid.length; k++){
			rMidFull[k] = rMid[k];
			rMidFull[rMid.length*2 - k - 1] = rMid[k];
			zMidFull[k] = zMid[k];
			zMidFull[zMid.length*2 - k - 1] = -zMid[k];
		}
		rMidFull[rMidFull.length-1] = rMidFull[0]; //close polygon
		zMidFull[zMidFull.length-1] = zMidFull[0];
		
		BinaryMatrixFile.mustWrite("/tmp/o.bin", new double[][]{ rOuterFull, zOuterFull }, true);
		BinaryMatrixFile.mustWrite("/tmp/i.bin", new double[][]{ rInnerFull, zInnerFull }, true);
		BinaryMatrixFile.mustWrite("/tmp/m.bin", new double[][]{ rMidFull, zMidFull }, true);
				
		double dPhi = 2* Math.PI / nCoils; 
			
		for(int i=0;i<nCoils;++i) {
			
			double R0 = 1.52; //1.65;
			double Z0 = 0;
			double a = 1.49; //1.62;
			double phi = i * dPhi;
			double w = 0.36;
			
			double cx = R0 * Math.cos(phi);
			double cy = R0 * Math.sin(phi);
			double cz = Z0;

			double ax = (R0+a) * Math.cos(phi);
			double ay = (R0+a) * Math.sin(phi);
			double az = Z0;
			
			double nx = -Math.sin(phi);
			double ny = Math.cos(phi);
			double nz = 0;
			
			System.out.println(cx+"\t"+ cy+"\t"+ cz+"\t"+ nx+"\t"+ ny+"\t"+ nz+"\t"+ ax+"\t"+ ay+"\t"+ az);
			
			//made up from diagrams
			
			Coil coil = new Coil();
			coil.setName("tfCoil_" + i);
			coil.setNumWindings(nWindingsPerCoil);
			coil.setWindingDirection(1);
			
			PolygonFilament poly = new PolygonFilament();
			
			double x[] = new double[rMidFull.length];
			double y[] = new double[rMidFull.length];
			for(int k=0; k < rMidFull.length; k++){
				x[k] = rMidFull[k] * FastMath.cos(phi);
				y[k] = rMidFull[k] * FastMath.sin(phi);
			}
			
			Points3D p = new Points3D();
			p.setX1(x);
			p.setX2(y);
			p.setX3(zMidFull.clone());
			poly.setVertices(p);
			coil.getCurrentCarrierPrimitive().add(poly);
			
			//double xyz[][] = n
			
						
			/*
			SolidCircularArc beam = new SolidCircularArc();
			beam.setName("tfArc_" + i);
			beam.setArcangle(2*Math.PI);
			
			beam.setNormal(vector(nx, ny, nz));
			beam.setR0(vector(ax, ay, az));
			beam.setCentre(vector(cx, cy, cz));
			
			beam.setHeight(w);
			beam.setWidth(w);
			coil.getCurrentCarrierPrimitive().add(beam);
			*/
			
			circuit.getCurrentCarrier().add(coil);
			
		}
		
		return configuration;
	}
	
	private static void tfFieldTest() {
		MagneticConfiguration conf = simpleTFModel();
		
		try {
			OneLiners.TextToFile("/tmp/tf.xml", Util.toXML(conf));
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		MagnetostaticsObjectModel magLib = new MagnetostaticsObjectModel(new RuntimeJava());
		MagnetostaticsResponsesLayer magCache = new MagnetostaticsResponsesLayer(new MagnetostaticsObjectModel(new RuntimeJava()));
		
		double Itf = -56e3; //56kA #30134 /sig/MAI/ITFsh, 55kA from /sig/MAI/ITFlam
		conf.getCircuit().get(0).setCurrent(Itf);
		
		double x[] = OneLiners.linSpace(-4.50, 4.50, 0.05);
		double y[] = OneLiners.linSpace(-4.00, 4.00, 0.05);
		
		double zz[] = OneLiners.fillArray(0, y.length);
		
		double Bx[][] = new double[x.length][];
		double By[][] = new double[x.length][];
		double Bz[][] = new double[x.length][];
		
		for(int i=0; i < x.length; i++){
			double xx[] = OneLiners.fillArray(x[i], y.length);
			double B[][] = magCache.magneticField(conf, new double[][]{ xx, y, zz} );
			//double B[][] = magLib.magneticField(conf, new double[][]{ xx, y, zz} );
			
			Bx[i] = B[0]; 
			By[i] = B[1]; 
			Bz[i] = B[2];
			
			System.out.println(i + " / " + x.length);
		}

		BinaryMatrixFile.mustWrite("/tmp/tfBx.bin", y, x, Bx, false);
		BinaryMatrixFile.mustWrite("/tmp/tfBy.bin", y, x, By, false);
		BinaryMatrixFile.mustWrite("/tmp/tfBz.bin", y, x, Bz, false);
		
		
		// 
		

	}
}

