package seed.minerva.magnetics.aug;

import java.util.Random;

import descriptors.aug.AUGSignalDesc;
import base.SignalFetcher;
import mds.AugMDSFetcher;
import mds.MdsPlusException;
import oneLiners.OneLiners;
import algorithmrepository.Interpolation1D;
import algorithmrepository.LinearInterpolation1D;
import seed.minerva.GraphicalModel;
import seed.minerva.deepend.Graph;
import seed.minerva.deepend.NodeImpl;
import seed.minerva.deepend.annotations.dependency;
import seed.minerva.deepend.annotations.dontSerialize;
import seed.minerva.diagnostics.ChannelDataEnable;
import signals.aug.AUGSignal;
import static seed.matrix.Mat.*;

/**
 * Beginnings of mag sensors for AUG
 * 
 * Signals we have:
 * /sig/MAD/Bthe[01:02:39] = B-theta-coil, measures poloidal magnetic field inside the vacuum vessel (in T)
 * /sig/MAG/Bthe[02:02:40] = B-theta-coil, measures poloidal magnetic field inside the vacuum vessel (in T)
 * /sig/MAG/Dpsi01D[01:16]D = differential flux, 2 loops 360 degrees inside the vacuum vessel (in Vs)
 * /sig/MAD/Br[1:10] = frame coil, measures radial magnetic field inside the vacuum vessel (in T)
 * /sig/MAS/Bthea[02:02:32] = B-theta-coil, measures poloidal magnetic field inside the vacuum vessel (in T)
 * /sig/MAS/Bthea[01:02:31] = B-theta-coil, measures poloidal magnetic field outside of the vacuum vessel (in T)
 * /sig/MAI/dpsia[01:17] = Flux outside of vessel coil 11 - coil 10  (?)
 * 
 */
public class PickupCoilsDataSourceAUG extends NodeImpl {
	public static String DEFAULTNAMEPREFIX = PickupCoilsDataSourceAUG.class.getSimpleName();
	
	// Parents
	@dependency private int pulse;
	@dependency private double time;
	@dependency boolean tfCompensationEnable = true;
	@dependency int[] enableRequest;	

	
	private ChannelDataEnable enable;
	
	private LinearInterpolation1D[] fullSignals;
	private LinearInterpolation1D bvacSignal;
	private double signalAtTime[];
	
	private String[] names;
	private double[] R, Z, angle, sigma, variance;
	private double tfResponse[];
			
	public PickupCoilsDataSourceAUG() { this(null, null, null, null); }
	public PickupCoilsDataSourceAUG(Graph g, String name) { this(g, name, null, null); }
	public PickupCoilsDataSourceAUG(Graph g, String name, Object pulse, Object time) {
		super(g, name);

		if(pulse != null)
			set("pulse", pulse);
		
		if(time != null)
			set("time", time);
	}
	
	private void loadSignalsOld(){
		SignalFetcher amds = AugMDSFetcher.defaultInstance();
		
		fullSignals = new LinearInterpolation1D[PickupsInfo.pickupInfoOld.length];
		R = new double[PickupsInfo.pickupInfoOld.length];
		Z = new double[PickupsInfo.pickupInfoOld.length];
		angle = new double[PickupsInfo.pickupInfoOld.length];
		names = new String[PickupsInfo.pickupInfoOld.length];
		
		sigma = new double[PickupsInfo.pickupInfoOld.length];
		variance = new double[PickupsInfo.pickupInfoOld.length];
		
		//Random r = new Random();
		for(int i=0; i < PickupsInfo.pickupInfoOld.length; i++){
			String signal = (String)PickupsInfo.pickupInfoOld[i][0];
			double geom[] = (double[])PickupsInfo.pickupInfoOld[i][1];
			R[i] = geom[0];
			Z[i] = geom[1];
			double normR = geom[2];
			double normZ = geom[3];
			names[i] = signal;
			
			angle[i] = Math.atan2(normR, -normZ ); //angle between what and what?
			
			//R[i] += 0.06 * (r.nextDouble() - 0.5);
			//Z[i] += 0.06 * (r.nextDouble() - 0.5);
			//angle[i] += 10 * Math.PI * (r.nextDouble() - 0.5) / 180;
			
			try{
				AUGSignal sig = (AUGSignal)amds.getSig("aug/" + pulse + signal);
				double data[] = (double [])sig.getDataAsType(double.class);
				double time[] = (double [])sig.getTVectorAsType(double.class);
				
				int n = 0;
				double avg = 0;
				for(; n < time.length; n++){
					if(time[n] > -10)
						break;
					avg += data[n];
				}
				avg /= n;
				
				for(int j=0; j < data.length; j++)
					data[j] -= avg;

				fullSignals[i] = new LinearInterpolation1D(time, data, Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
				
				sigma[i] = ((double[])PickupsInfo.pickupInfoOld[i][1])[4];
				variance[i] = sigma[i]*sigma[i];
				
			}catch(Exception e){
				System.err.println("AUG pickups: Unable to load '"+signal+"'.");
				fullSignals[i] = null;
			}
			
		}

		signalAtTime = null;
		if(enable == null)
			enable = new ChannelDataEnable(PickupsInfo.pickupInfoOld.length);
	}
	
	private void loadSignals(){
		SignalFetcher amds = AugMDSFetcher.defaultInstance();
		
		int n = PickupsInfo.pickups.length;
		fullSignals = new LinearInterpolation1D[n];
		R = new double[n];
		Z = new double[n];
		angle = new double[n];
		names = new String[n];
		
		sigma = new double[n];
		variance = new double[n];
		
		//Random r = new Random();
		for(int i=0; i < n; i++){
			String signal = PickupsInfo.pickups[i][0];			
			R[i] = Double.parseDouble(PickupsInfo.pickups[i][1]);
			Z[i] = Double.parseDouble(PickupsInfo.pickups[i][2]);
			names[i] = signal;
			
			angle[i] =  Double.parseDouble(PickupsInfo.pickups[i][3]); //angle between what and what?
			
			//R[i] += 0.06 * (r.nextDouble() - 0.5);
			//Z[i] += 0.06 * (r.nextDouble() - 0.5);
			//angle[i] += 10 * Math.PI * (r.nextDouble() - 0.5) / 180;
			
			try{
				AUGSignal sig = (AUGSignal)amds.getSig("aug/" + pulse + signal);
				double data[] = (double [])sig.getDataAsType(double.class);
				double time[] = (double [])sig.getTVectorAsType(double.class);
				
				int nt = 0;
				double avg = 0;
				for(; nt < time.length; nt++){
					if(time[nt] > -10)
						break;
					avg += data[nt];
				}
				avg /= nt;
				
				for(int j=0; j < data.length; j++)
					data[j] -= avg;

				fullSignals[i] = new LinearInterpolation1D(time, data, Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
				
				sigma[i] = PickupsInfo.defaultSigma; //filled later by probe test results
				variance[i] = sigma[i]*sigma[i];
				
			}catch(Exception e){
				System.err.println("AUG pickups: Unable to load '"+signal+"'.");
				fullSignals[i] = null;
			}
			
		}
		
		AUGSignal BVACSig = (AUGSignal)amds.getSig("/aug/" + pulse + "/sig/MAI/BTF");		
		double btfSig[] = (double[])BVACSig.getDataAsType(double.class);
		double btfTime[] = (double[])BVACSig.getTVectorAsType(double.class);
		bvacSignal = new LinearInterpolation1D(btfTime, btfSig, Interpolation1D.EXTRAPOLATE_EXCEPTION, 0);
		

		signalAtTime = null;
		if(enable == null)
			enable = new ChannelDataEnable(n);
	}
	
	private void updateSignals(){
		int[] status = enable.getStatus();
		
		double btf = 0;
		if(tfCompensationEnable){
			btf = bvacSignal.eval(time);
		}
		
		signalAtTime = new double[fullSignals.length];
		for(int i=0; i < fullSignals.length; i++){
			if(fullSignals[i] != null){
				signalAtTime[i] = fullSignals[i].eval(time);
				status[i] = ChannelDataEnable.DATA_VALID;
				
				if(tfCompensationEnable){
					signalAtTime[i] -= tfResponse[i] * btf;
				}
			} else {
				signalAtTime[i] = Double.NaN;
				sigma[i] = Double.NaN;
				variance[i] = Double.NaN;
				status[i] = ChannelDataEnable.DATA_NOT_PRESENT;
			}
		}
		
		enable.setStatus(status);
		enable.updateEnable();
	}
	
	
	/** Output of aug_probe_test, using probe test to identify defective probes and estimate sigmas */
	private void loadProbeTest(){		
		String ptNames[];
		double pTSigma[];
		double alphaTF[];
		
		int enReq[] = OneLiners.fillIntArray(ChannelDataEnable.OFF, enable.getNumChannels());
		tfResponse = new double[enable.getNumChannels()];
		
		try{
			AUGSignalDesc desc = new AUGSignalDesc(pulse, "probeTest", "names");		
			AUGSignal sig = (AUGSignal) AugMDSFetcher.defaultInstance().getSig(desc);
			ptNames = (String[])sig.getData();
			
			desc = new AUGSignalDesc(pulse, "probeTest", "sigma");		
			sig = (AUGSignal) AugMDSFetcher.defaultInstance().getSig(desc);
			pTSigma =  (double[])sig.getData();
			
			desc = new AUGSignalDesc(pulse, "probeTest", "alphaTF");		
			sig = (AUGSignal) AugMDSFetcher.defaultInstance().getSig(desc);
			alphaTF =  (double[])sig.getData();		
			
		}catch(RuntimeException err){
			System.err.println("WARNING: No probe test sigma information: " + err.getMessage() +". *** ALL SENSORS WILL BE DISABLED! ***");
			return;
		}
		
		for(int i=0; i < ptNames.length; i++){
			if(Double.isNaN(pTSigma[i])){
				System.out.println("Pickup '"+ptNames[i]+"' disabled by probe test failure.");
				continue;
			}
			
			int idx = getSensorIndexInternal(ptNames[i]);
			if(idx < 0){
				System.err.println("WARNING: Probe test sigma present for unknown signal: " + ptNames[i]);
				continue;
			}
			
			enReq[idx] = ChannelDataEnable.ON;
			tfResponse[idx] = alphaTF[i];
			sigma[idx] = pTSigma[i];
			variance[idx] = sigma[idx]*sigma[idx];			
		}
		
		enable.setEnableRequest(enReq);
	}
	
	
	/*
	public void setEnableRequest(int[] enableRequest) {
		update(); //we need the enable object to exist fist
		
		if(enableRequest.length != names.length)
			throw new IllegalArgumentException("setEnableRequest(en[" + enableRequest.length +"] but there are "+names.length+" sensors.");
		
		
		enable.setEnableRequest(enableRequest);
		enable.updateEnable();
		
		setChanged("enable");
	}
	
	public void setEnableRequest(int enableRequestForAll) {
		update();
		
		enable.setEnableRequest(OneLiners.fillIntArray(enableRequestForAll, enable.getNumChannels()));
		enable.updateEnable();
		
		setChanged("enable");
	}
	
	public int[] getEnableRequest() { update(); return enable.getEnableRequest(); }
	
	public void setEnableRequest(int index, int request) {
		update(); //we need the enable object to exist first
		int i = 0;
		int[] enableRequest = enable.getEnableRequest();
		enableRequest[index] = request;
		enable.setEnableRequest(enableRequest);
		enable.updateEnable();
		
		setChanged("enable");
	}

	
	public boolean[] getEnable() {
		update();
		return enable.getEnable();
	}
	
	*/
	
	@dontSerialize
	public void setEnableRequest(Object enableRequest) {
		set("enableRequest", enableRequest);		
	}
	
	public int[] getEnableRequest() {
		update("enableRequest");
		return enableRequest;
	}
	
	public void setEnableRequest(int enableRequestForAll) {
		update();
		setEnableRequest(OneLiners.fillIntArray(enableRequestForAll, enable.getNumChannels()));		
	}
	
	public void setEnableRequest(int index, int request) {
		update();
		int[] newEnableRequest = null;
		if (!isSet("enableRequest") || enableRequest == null) {
			newEnableRequest = enable.getEnableRequest();
		} else {
			newEnableRequest = enableRequest;
		}				
		
		newEnableRequest[index] = request;
		setEnableRequest(newEnableRequest);		
	}
	
	
	public void setEnableRequest(String[] name, int[] request) {
		for(int i=0;i<name.length;++i) {
			setEnableRequest(name[i],request[i]);
		}
	}	
	
	public boolean[] getEnable() {
		update();
		return enable.getEnable();
	}

	
	public double[] getR() { update();return R; }	
	public double[] getZ() { update();return Z; }	
	public double[] getAngle() { update();return angle;	}
	public double[] getPoloidalAngle() { update();return angle;	}
	public String[] getNames() { update();return names;	}	
	public double[] getData() {	update(); return signalAtTime; }
	
	public double[] getBiasMean(){ update(); return new double[names.length]; }
	public double[] getBiasVariance(){ update(); return OneLiners.fillArray(1, names.length); }
	public double[] getVariance(){ update(); return variance; }
	
	
	@Override
	public void updateState() {
		// Invalidate
		if (fullSignals == null || enable == null || isChanged("pulse")) {
			loadSignals();
			loadProbeTest();
		}
		
		
		if (isChanged("enableRequest")) {
			enable.setEnableRequest(enableRequest);
			enable.updateEnable();			
		}
		
		
		if (signalAtTime == null || isChanged("time") || isChanged("tfCompensationEnable")) {
			updateSignals();
		}
		
		
	}
	
		@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		names = null;
		R = null; Z = null; angle = null;
		sigma = null; variance = null;
		fullSignals = null;
		signalAtTime = null;
	}
	
	public int getSensorIndex(String name){
		update();
		return getSensorIndexInternal(name);
	}
	
	private int getSensorIndexInternal(String name){
		for(int i=0; i < names.length; i++){
			if(name.equalsIgnoreCase(names[i])){
				return i;
			}
		}
		return -1;
	}

	public void setEnableRequest(String name, int request) {
		setEnableRequest(getSensorIndex(name), request);
	}

	public void setTFCompensationEnable(boolean tfCompensationEnable) { 
		set("tfCompensationEnable", tfCompensationEnable);
	}

}
