package seed.minerva.neutralBeams.aug;

import jafama.FastMath;
import oneLiners.OneLiners;
import algorithmrepository.Algorithms;
import seed.minerva.neutralBeams.GaussianBeamDefinition;

/** Based on the data in AugHeatingBeams, provide multiple beamlets
 * to improve geometry modelling.
 * 
 * For a number of points on the source, a number of beamlets to
 * different points in the beam focus are calculated.  
 *  
 */
public class AugHeatingBeamlets extends AugHeatingBeams {
	
	/** config of points on beam source */
	private double sourceWidth = 0.228, sourceHeight = 0.506;
	private int nSourcePointsInWidth = 5;//3;
	private int nSourcePointsInHeight = 5;//4;
	
	/** Vertical vector of the beam box. (usually global up) */
	private double boxUp[] = new double[]{ 0, 0, 1 };
	
	/** Distance (forwards) along the beam line from the source to the focus.
	 * This is used to find the source position, by tracing back up the beamline */
	private double sourceVertFocusDistance = 8.5;
	
	private double focusGridWidth = 0.30, focusGridHeight = 0.30;
	private double focusFWHMWidth = 0.11, focusFWHMHeight = 0.15; 
	//private int nFocusPointsWidth = 5, nFocusPointsHeight= 6;
	private int nFocusPointsWidth = 8, nFocusPointsHeight= 8;
	
	/** Size of beamlets (FWHM at focus) as fraction of the beamlet spacing at focus */
	private double beamletFWHMFractionOfSpacing = 1.7;
	
	private GaussianBeamDefinition fullBeamDefs[];
	
	@Override
	public void calcBeams() {
		super.calcBeams();
		fullBeamDefs = beamDefs;
		
		double fwhmToSigma = 1.0 / 2.35;
		
		int nBeamletsPerSource = nSourcePointsInWidth * nSourcePointsInHeight * nFocusPointsWidth * nFocusPointsHeight;		
		GaussianBeamDefinition beamlets[] = new GaussianBeamDefinition[fullBeamDefs.length * nBeamletsPerSource];
		
		for(int iF=0; iF < fullBeamDefs.length; iF++){ //for full source (in the original defs)			
			double sourcePos[] = OneLiners.minus(fullBeamDefs[iF].focus, OneLiners.mul(fullBeamDefs[iF].uVec, sourceVertFocusDistance));
			double sourceRight[] = OneLiners.reNorm(OneLiners.cross(fullBeamDefs[iF].uVec, boxUp));
			double sourceUp[] = OneLiners.reNorm(OneLiners.cross(sourceRight, fullBeamDefs[iF].uVec));
			
			//double sigmaAtFocusSq = FastMath.pow2(fullBeamDefs[iF].fwhmAtFocus / 2.35);
			
			//calculate fraction for beamlet distributions at focus (assume its the same for all source points) 
			double beamFrac[][] = new double[nFocusPointsHeight][nFocusPointsWidth];
			double beamFracSum = 0;
			for(int iFY = 0; iFY < nFocusPointsHeight; iFY++){
				for(int iFX = 0; iFX < nFocusPointsWidth; iFX++){
					double fx = (((double)iFX / (nFocusPointsWidth - 1.0)) - 0.5) * focusGridWidth;
					double fy = (((double)iFY / (nFocusPointsHeight - 1.0)) - 0.5) * focusGridHeight;
					double r2 =   (fx*fx) / FastMath.pow2(focusFWHMWidth * fwhmToSigma) 
							     + (fy*fy) / FastMath.pow2(focusFWHMHeight * fwhmToSigma);
					beamFrac[iFY][iFX] = FastMath.exp( -r2 / 2);
					
					beamFracSum += beamFrac[iFY][iFX];
				}
			}
			beamFracSum *= (nSourcePointsInWidth * nSourcePointsInHeight); //totals are for all source and all focus points
			
			
			//foreach source point
			for(int iSY = 0; iSY < nSourcePointsInHeight; iSY++){
				for(int iSX = 0; iSX < nSourcePointsInWidth; iSX++){
					double sx = (((double)iSX / (nSourcePointsInWidth - 1.0)) - 0.5) * sourceWidth;
					double sy = (((double)iSY / (nSourcePointsInHeight - 1.0)) - 0.5) * sourceHeight;
					
					double beamletSrcPos[] = new double[]{
							sourcePos[0] + sx * sourceRight[0] + sy * sourceUp[0],
							sourcePos[1] + sx * sourceRight[1] + sy * sourceUp[1],
							sourcePos[2] + sx * sourceRight[2] + sy * sourceUp[2],
					};
					
					//foreach focus point
					for(int iFY = 0; iFY < nFocusPointsHeight; iFY++){
						for(int iFX = 0; iFX < nFocusPointsWidth; iFX++){
							//beamlet index
							int iBL = ((((((iF * nSourcePointsInHeight) + iSY) * nSourcePointsInWidth) + iSX) * nFocusPointsHeight) + iFY) * nFocusPointsWidth + iFX;
							
							beamlets[iBL] = new GaussianBeamDefinition();
							
							double fx = (((double)iFX / (nFocusPointsWidth - 1.0)) - 0.5) * focusGridWidth;
							double fy = (((double)iFY / (nFocusPointsHeight - 1.0)) - 0.5) * focusGridHeight;
							beamlets[iBL].focus = new double[]{
									fullBeamDefs[iF].focus[0] + fx * sourceRight[0] + fy * sourceUp[0],
									fullBeamDefs[iF].focus[1] + fx * sourceRight[1] + fy * sourceUp[1],
									fullBeamDefs[iF].focus[2] + fx * sourceRight[2] + fy * sourceUp[2],
							};							
							
							beamlets[iBL].uVec = OneLiners.reNorm(OneLiners.minus(beamlets[iBL].focus, beamletSrcPos));
							//make FWHM at focus same as avg. dist between beamlets
							beamlets[iBL].fwhmAtFocus = (  ((double)focusGridWidth / (nFocusPointsWidth - 1.0)) 
									                       + ((double)focusGridHeight / (nFocusPointsHeight - 1.0)) ) / beamletFWHMFractionOfSpacing;
							beamlets[iBL].fwhmIncrease = 0.0001;//fullBeamDefs[iF].fwhmIncrease; //same divergence as overall beam
							beamlets[iBL].velocityAngularVariance = fullBeamDefs[iF].velocityAngularVariance;
							beamlets[iBL].atomicNo = fullBeamDefs[iF].atomicNo;
							beamlets[iBL].atomicMass = fullBeamDefs[iF].atomicMass;
							beamlets[iBL].cmpInjectionRate = fullBeamDefs[iF].cmpInjectionRate.clone();
							beamlets[iBL].cmpPowerFrac = fullBeamDefs[iF].cmpPowerFrac.clone();
							beamlets[iBL].cmpNumberFrac = fullBeamDefs[iF].cmpNumberFrac.clone();
							beamlets[iBL].cmpEnergyMean = fullBeamDefs[iF].cmpEnergyMean.clone();
							beamlets[iBL].cmpEnergyVar = fullBeamDefs[iF].cmpEnergyVar.clone();
							
							for(int iC=0; iC < beamlets[iBL].cmpInjectionRate.length; iC++){
								beamlets[iBL].cmpInjectionRate[iC] *= beamFrac[iFY][iFX] / beamFracSum;
								beamlets[iBL].cmpPowerFrac[iC] *= beamFrac[iFY][iFX] / beamFracSum;
								beamlets[iBL].cmpNumberFrac[iC] *= beamFrac[iFY][iFX] / beamFracSum;								
							}
						}
					}
				}	
			}	
		}
				
		beamDefs = beamlets;
	}
	
	public void setSourceResolution(int nSourcePointsInWidth, int nSourcePointsInHeight){
		this.nSourcePointsInWidth = nSourcePointsInWidth;
		this.nSourcePointsInHeight = nSourcePointsInHeight;
		beamDefs = null;
		setChanged();
	}
	
	public void setFocusResolution(int nFocusPointsWidth, int nFocusPointsHeight){
		this.nFocusPointsWidth = nFocusPointsWidth;
		this.nFocusPointsHeight = nFocusPointsHeight;
		beamDefs = null;
		setChanged();
	}
	
	public GaussianBeamDefinition[] getFullBeamDefinitions(){ update(); return fullBeamDefs; }	

}
