package seed.minerva.neutralBeams.aug;


import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import mds.AugMDSFetcher;
import base.SignalFetcher;
import descriptors.aug.AUGSignalDesc;
import otherSupport.RandomManager;
import algorithmrepository.Algorithms;
import jafama.FastMath;
import oneLiners.OneLiners;
import seed.minerva.deepend.Graph;
import seed.minerva.deepend.NodeImpl;
import seed.minerva.deepend.annotations.dependency;
import seed.minerva.neutralBeams.BeamDefsNode;
import seed.minerva.neutralBeams.GaussianBeamDefinition;
import seed.minerva.neutralBeams.GaussianBeamDefinitions;
import signals.aug.AUGSignal;

/** Geometry and fuel information for the ASDEX Upgrade Neutral Beams
 * 
 *  Much of this is the geometry definitions from the AUG NBI website
 at http://www.ipp.mpg.de/technologie/nbi/AUG/geometry/

 The only modification is that we define theta as always the angle to north
 and currently have +ve x as North and +ve y as West.

 It starts with the definition of a point P0 (R=R_P0, phi=theta) 
   which happens to be the common focus in the horizonal of each pair
   of beams (but doesn't have to be here)
   
 The beam box axis lies at  angle alpha to the radial in the (x,y) plane, through P0.
 The beam axis lies at angle psi to the beam axis in the (x,y) plane.
 So the beam line lies at alpha+psi to the radial in the (x,y) plane.
 
 A distance Dx from P0 along the beam line is where the beam line goes through Z=0
 
 The angle between Z=0 and the beam line in the (R,Z) plane is beta. 
 So the point A = P0 + [ deltaX * tan(beta) _y^ ] is a point, on the beam line, directly above P0
 
 The point Px is then Px = P0 + deltaX * [cos(alpha + psi - theta), -sin(alpha + psi - theta), 0]
 
 
 Much of it though, it completely made up, but is marked as such.  
 */
public class AugHeatingBeams extends NodeImpl implements GaussianBeamDefinitions, BeamDefsNode {
	public static String DEFAULTNAMEPREFIX = AugHeatingBeams.class.getSimpleName();
	
	public final static double electronCharge = 1.60217646e-19;
	public static final double protonMass = 1.67262158e-27; //kg
	private final static double[] deg2Rad(double deg[]){ for(int i=0; i < deg.length; i++){ deg[i] *= Math.PI / 180; } return deg; }
	private final static double[][] deg2Rad(double deg[][]){ for(int i=0; i < deg.length; i++){ deg[i] = deg2Rad(deg[i]); } return deg; }
	
	public static final int BEAMBOX_SE = 0; //Southeast was always the same
	public static final int BEAMBOX_NW2001 = 1; //(05/2001+ Current Drive)
	public static final int BEAMBOX_NW1998 = 2; //(12/1998-6/2000)
	public static final int BEAMBOX_NW1997 = 3; //(12/1997-8/1998)
	
	public static final int BEAMFUEL_HYDROGEN = 0;
	public static final int BEAMFUEL_DEUTERIUM = 1;
	
	/** Time index for beam auto selection. (Optional) */
	@dependency private int pulse;
	/** Time index for beam auto selection. (Optional) */
	@dependency private double time;  
	
	/** Fuel to use when beams are auto selected (Default is BEAMFUEL_DEUTERIUM)*/
	@dependency private int autoSelectFuel; 
	
	/** Fraction of power in each beam component [BEAMBOX_*][BEAMFUEL_*][component] */
	@dependency private double[][][] componentPowerFractions;
	
	/** Transmission (injected / extracted)  current/power [BEAMBOX_*][BEAMFUEL_*]  */
	@dependency private double transmission[][];

	/** Spread of each component as fraction of calculated energy
	(i.e the 1/2 component is expected to have 1/2 the deviation of the full) 
	FIXME: made up to match Reimer's spectra
	*Set in constructor */	
	@dependency private double energyFractionalStdDeviation;  
	
	/** FWHM of beam velocity angular spread (at any point). 
	 * FIXME: made up to match Reimer's spectra
			*Set in constructor */
	@dependency private  double velocityAngularVariance;
	
	@dependency private int selectedBeams[][];
			
	/** Deposited power signals used only for beam selection */	
	private static final String powerSignal = "/sig/NIS/PNIQ";
	private static final double beamOnPowerThreshold = 100e3; //100kW
	
	
	/* ******** Beam geometry tweaked to match scorch marks on innner heat shield from faroNov2013 ***** */
	/** If true, beams are steered (keeping PINI positions), to match the burn marks
	 * on the wall measured by FARO (nov2013) */
	@dependency private boolean matchGeomToNov2013Faro = true;
	private double sourcePivotR = 7.6; //very roughly, from a PDF diagram of R,Z projection of the box
	private double vesselContactPointsFaroNov2013[][][] = new double[][][]{
			{
				{ -0.623, -0.845, -0.115 }, //Q1
				//{ -0.269, -1.017, -0.086 }, //Q2 as measured
				{ -0.266, -1.016, -0.175 }, //Q2 on the bolt below where measured, so assuming we put the tile back wrong
				{ -0.237, -1.022, 0.147 }, //Q3: actually just the bit of string setup on Q3's theory axis - not really a measurement
				{ -0.603, -0.858, 0.134 }, //Q4
			}, 
			{ }, // Q5-8, no measurements 
			{}, {}, //old Q5-8 no measurements 
	};
	
	/* ******** Beam geometry from NBI group specifications ***** */
	
	// Geometry is defined for each  BEAMBOX_* and for each of te 4 beams within a beam box
	
	/** Beam inclination angle */
	private static final double beta[][] = deg2Rad(new double[][]{ 
				{ 4.9, 4.90, -4.90, -4.90 }, 		
				{ 4.9, 6.65, -6.65, -4.90 },
				{}, {} /* Not filled yet */
			});
	
	/** Angle of R to south/north */	
	private static final double theta[][] = deg2Rad(new double[][]{ 
				{ 180+33.75, 	180+33.75, 	180+33.75, 	180+33.75 },	
				{ 29.0, 		29.0, 		29.0, 		29.0 },
				{}, {} /* Not filled yet */
			}); 
	
	/** Angle FROM the reverse box axis (opposite to injection direction)
	 * 	TO R (radial from (0,0) to P0). +ve for anticlockwise */
	private static final double psi[][] = deg2Rad(new double[][]{ 
				{ 15.0, 15.0, 15.0, 15.0 },		
				{ 18.9, 18.9, 18.9, 18.9 },
				{}, {} /* Not filled yet */
			});
	
	 /** Horizontal angle of beam axis to box axis */
	private static final double alpha[][] = deg2Rad(new double [][]{ 
				{ -4.1, 4.1, 4.1, -4.1 },		
				{ -4.1, 4.1, 4.1, -4.1 },
				{}, {} /* Not filled yet */
			});
	
	 /** Radial position of horzontal beam crossing */
	private static final double R_P0[][] = new double[][]{ 
				{ 2.842, 2.842, 2.842, 2.842 },	
				{ 3.296, 3.296, 3.296, 3.296 },
				{}, {} /* Not filled yet */
			};
	
	/** Distance of Z=0 crossing along box axis from P0 
	 * +ve is toward the beam source / away from R=0 */ 
	private static final double Dx[][] = new double[][]{ 
				{ -0.5, -0.5, -0.5, -0.5 }, 		
				{ -0.5,  0.5,  0.5, -0.5 },
				{}, {}/* Not filled yet */
			}; 
	
	/** The beam indecies actually match the AUG naming convention (except for the 1-based indexing)
	 * (select from the appropriate beam box to select between e.g. Q3 and Q7) 
	 * I think Q means 'Quelle' ('source' in German) */  
	public static final int BEAM_Q1_Q5 = 0;
	public static final int BEAM_Q2_Q6 = 1;
	public static final int BEAM_Q3_Q7 = 2;	
	public static final int BEAM_Q4_Q8 = 3;

	/** The standard MSE/CXRS etc beam is Q3 - SW upper tangential*/
	public static final int BEAM_MSE_STD = BEAM_Q3_Q7;
	
	
	// Some geometry is just all the same (mostly stuff we made up for now)
	/** FWHM of beam at focus (roughly, from http://dx.doi.org/10.1016/0010-4655(95)00013-6 ) 
	 * A.Herrmann: Asdex Upgrade Grfäßphotos suggests much smaller
	 * NBI group modelling (DENSB) suggests 30cm half width at foot (0.1 of max)
	 * Tentitativly: IMSE data suggests smaller too. 
	 */
	private static final double fwhmAtFocus = 0.16;
	//private static final double fwhmAtFocus = 0.10;
	
	/** Increase in FWHM per unit length along beam axis i.e. tanTheta 
	 * This is a crude estimate, based on the size of extraction grids (50.6 x 22.8 cm) at 
	 * the distance from them to the vertical (8.5m) and horizontal (6.5m) focal points.
	 * 
	 * It's also vaguely consistent with 'about 1 degree' min divergence on the NBI website.
	 * 
	 *  IMSE image suggests this is very small.
	 *  Changing it seems to have no effect on the predicted image though.
	 * */ 
	private static final double fwhmIncrease = 0.8 * Math.PI / 180; //assume entirely from beamLET divergence
	//private static final double fwhmIncrease = 0.02; 
		
	/** FWHM of beam velocity angular spread (at any point). 
	 * FIXME: made up to match Reimer's spectra
			*Set in constructor */
	private static double defaultVelocityAngularVariance = 1.5*1.5 *Math.PI*Math.PI /180/180;

	/** Spread of each component as fraction of calculated energy
			(i.e the 1/2 component is expected to have 1/2 the deviation of the full) 
			FIXME: made up to match Reimer's spectra
			*Set in constructor */	
	private static double defailtEnergyFractionalStdDeviation = 0.015;  
	
	// Fuel and acceleration info is stored for D and H, and also for each of [BEAMBOX_*][BEAMFUEL_*]
	// These could in principal come from sig/NIS/INJ1:xxx
	// See http://www.ipp.mpg.de/technologie/nbi/AUG/NIS.html
	
	/** Atomic No and mass per fuel type [BEAMFUEL_*] */
	private static final int atomicNo[] = { 1, 1 };
	private static final int atomicMass[] = { 1, 2 };
	
	/** Beam accelerating voltage in V [BEAMBOX_*][BEAMFUEL_*] */
	private static final double accelVoltage[][] = new double[][]{ { 55e3, 60e3 }, { 72e3, 93e3 }, {Double.NaN,Double.NaN}, {Double.NaN,Double.NaN} };

	/** Perveance at optimum in AV^(-3/2) [BEAMBOX_*][BEAMFUEL_*]  */
	private static final double optimumPerveance[][] = new double[][]{ { 7.0e-6, 5.3e-6 }, { 3.2e-6, 2.2e-6}, {Double.NaN,Double.NaN}, {Double.NaN,Double.NaN} };
	
	/** Transmission (injected / extracted)  current/power [BEAMBOX_*][BEAMFUEL_*]  */
	public final static double defaultTransmission[][] = new double[][]{ { 0.363, 0.525 }, { 0.321, 0.423 }, {Double.NaN,Double.NaN}, {Double.NaN,Double.NaN} };

	/** Fraction of power in each beam component [BEAMBOX_*][BEAMFUEL_*][component] */
	public final static double defaultComponentPowerFractions[][][] = new double[][][]{
			{ { 0.50, 0.30, 0.20 },  { 0.65, 0.25, 0.10 } },  // SE: H, D
			{ { 0.38, 0.33, 0.29 },  { 0.64, 0.26, 0.10 } }, // NW2001+: H, D
			{ {Double.NaN,Double.NaN}, {Double.NaN,Double.NaN} },
			{ {Double.NaN,Double.NaN}, {Double.NaN,Double.NaN} },
		};
	
	/** Molecular mass of components when accelerated (i.e. before disassociation) */
	private int componentMolecularMass[] = new int[]{ 1, 2, 3 };
	
	
	/** Beam setup for today, [beamIdx][type/geomIdx/fuel atomic no.] */	
	public final static int defaultSelectedBeams[][] = new int[][]{ 
			{ BEAMBOX_SE, BEAM_Q1_Q5, BEAMFUEL_DEUTERIUM }, //NORM LOW
			{ BEAMBOX_SE, BEAM_Q2_Q6, BEAMFUEL_DEUTERIUM }, //TANG LOW
			{ BEAMBOX_SE, BEAM_Q3_Q7, BEAMFUEL_DEUTERIUM },  //TANG HIGH (MSE)
			{ BEAMBOX_SE, BEAM_Q4_Q8, BEAMFUEL_DEUTERIUM }, //NORM HIGH
			{ BEAMBOX_NW2001, BEAM_Q1_Q5, BEAMFUEL_DEUTERIUM }, // NORM LOW
			{ BEAMBOX_NW2001, BEAM_Q2_Q6, BEAMFUEL_DEUTERIUM }, // TANG LOW (Current Drive)
			{ BEAMBOX_NW2001, BEAM_Q3_Q7, BEAMFUEL_DEUTERIUM }, // TANG HIGH (Current Drive)
			{ BEAMBOX_NW2001, BEAM_Q4_Q8, BEAMFUEL_DEUTERIUM }, // NORM HIGH
	};	
	
	protected GaussianBeamDefinition beamDefs[];
		
	public AugHeatingBeams() { this(null, null, null, null); }
	public AugHeatingBeams(Graph g, String name) { this(g, name, null, null); }
	public AugHeatingBeams(Graph g, String name, Object pulse, Object time) { 
		super(g, name);
		
		set("velocityAngularVariance", defaultVelocityAngularVariance);
		set("energyFractionalStdDeviation", defailtEnergyFractionalStdDeviation);
		set("componentPowerFractions", defaultComponentPowerFractions);
		set("transmission", defaultTransmission);
		set("autoSelectFuel", BEAMFUEL_DEUTERIUM);
		set("selectedBeams", defaultSelectedBeams);
				
		if(pulse != null)
			set("pulse", pulse);
		if(time != null)
			set("time", time);
	}
	
	public void calcBeams() {
		if(componentPowerFractions == null){
			System.err.println("WARNING: Lost componentPowerFractions, resetting to defaults (probably minerva3 bug)");
			componentPowerFractions = defaultComponentPowerFractions;
		}
		
		int nBeams = selectedBeams.length;
		
		beamDefs = new GaussianBeamDefinition[nBeams];
		
		for(int i=0; i < nBeams; i++){
			beamDefs[i] = new GaussianBeamDefinition();
			
			int type = selectedBeams[i][0];
			int geom = selectedBeams[i][1];
			int fuel = selectedBeams[i][2];
			
			//first sort out the geometry
			double A[] = new double[3];
			double Px[] = new double[3];
			double u[] = new double[3];
			
			A[0] = R_P0[type][geom] * Math.cos(theta[type][geom]);
			A[1] = R_P0[type][geom] * Math.sin(theta[type][geom]);
			A[2] = Dx[type][geom] * Math.tan(-beta[type][geom]);
			
			Px[0] = A[0] + Dx[type][geom] * Math.cos(theta[type][geom] - psi[type][geom] - alpha[type][geom]);
			Px[1] = A[1] + Dx[type][geom] * Math.sin(theta[type][geom] - psi[type][geom] - alpha[type][geom]);
			Px[2] = 0;
			
			double l = Dx[type][geom] / Math.cos(-beta[type][geom]); //length between A and Px, by def.
			
			u[0] = (A[0] - Px[0]) / l;
			u[1] = (A[1] - Px[1]) / l;
			u[2] = (A[2] - Px[2]) / l;
			
			beamDefs[i].uVec = u;
			
			//FIXME: Where actually is the focus? Here we assume it conincides with the beam crossing focus
			beamDefs[i].focus = Px; 
			
			if(matchGeomToNov2013Faro && vesselContactPointsFaroNov2013[type].length > geom)
				matchGeomToNov2013Faro(i, vesselContactPointsFaroNov2013[type][geom]);
			
			beamDefs[i].fwhmAtFocus = fwhmAtFocus;
			beamDefs[i].fwhmIncrease = fwhmIncrease;
			beamDefs[i].velocityAngularVariance = velocityAngularVariance;
			
			//and now the fuel/accel info
			
			beamDefs[i].atomicNo = atomicNo[fuel];
			beamDefs[i].atomicMass = atomicMass[fuel];

			double extractedCurrent = optimumPerveance[type][fuel] * Math.pow(accelVoltage[type][fuel], 3.0/2.0);

			//molecules per second leaving the extraction grid
			double moleculeExtractionRate = extractedCurrent / (atomicNo[fuel] * electronCharge); //number of full energy particles (molecules or atoms) per beam

			int nComponents = componentMolecularMass.length;
			
			beamDefs[i].cmpPowerFrac = componentPowerFractions[type][fuel];
			beamDefs[i].cmpNumberFrac = new double[nComponents];
			beamDefs[i].cmpEnergyMean = new double[nComponents];
			beamDefs[i].cmpEnergyVar = new double[nComponents];
			beamDefs[i].cmpInjectionRate = new double[nComponents];
			
			// Calc the fraction of the injected atoms in each component
			double sumNumFrac = 0;			
			for(int iC=0; iC < nComponents; iC++) {
				double energyFrac = 1.0 / componentMolecularMass[iC];
				beamDefs[i].cmpNumberFrac[iC] = beamDefs[i].cmpPowerFrac[iC] / energyFrac;
				sumNumFrac += beamDefs[i].cmpNumberFrac[iC];
			}
				
			for(int iC=0; iC < nComponents; iC++) {
				beamDefs[i].cmpNumberFrac[iC] /= sumNumFrac;
				
				double energyFrac = 1.0 / componentMolecularMass[iC];
				
				beamDefs[i].cmpEnergyMean[iC] = accelVoltage[type][fuel] * energyFrac;
				beamDefs[i].cmpEnergyVar[iC] = (energyFractionalStdDeviation * beamDefs[i].cmpEnergyMean[iC]) 
												* (energyFractionalStdDeviation * beamDefs[i].cmpEnergyMean[iC]);
				
				// number of individual atoms injected
				double atomExtractionRate = moleculeExtractionRate * beamDefs[i].cmpPowerFrac[iC] * componentMolecularMass[iC]; 

				beamDefs[i].cmpInjectionRate[iC] = atomExtractionRate * transmission[type][fuel];

			}
		}
		
	}
	
	/** Move beam axes to match Faro Nov2013 measurements of burn marks */
	private void matchGeomToNov2013Faro(int i, double vesselContactPos[]) {
		if(i >= vesselContactPointsFaroNov2013.length)
			return;
		
		//at this point beamDefs[i].uVec and .focus have been filled in
		
		//save R of focus for later
		double focusR2 = FastMath.pow2(beamDefs[i].focus[0]) + FastMath.pow2(beamDefs[i].focus[1]);
		
		//find point along that line of source pivot
		double t[] = Algorithms.cylinderLineIntersection(beamDefs[i].focus, beamDefs[i].uVec, 
										new double[]{ 0, 0, 0 },
										new double[]{ 0, 0, 1 },
										sourcePivotR*sourcePivotR);
		
		double srcPos[] = OneLiners.plus(beamDefs[i].focus, OneLiners.mul(beamDefs[i].uVec, Algorithms.min(t))); //most -ve on line
		
		//new unit vec from source to vessel contact point
		beamDefs[i].uVec = OneLiners.reNorm(OneLiners.minus(vesselContactPos, srcPos));
				
		//and work out the focus at the old R, along the new line
		t = Algorithms.cylinderLineIntersection(srcPos, beamDefs[i].uVec, 
				new double[]{ 0, 0, 0 },
				new double[]{ 0, 0, 1 },
				focusR2);
		
		beamDefs[i].focus = OneLiners.plus(srcPos, OneLiners.mul(beamDefs[i].uVec, Algorithms.min(t)));
		
	}
	
	private void updateBeamSelection() {
		if(pulse < 0 || time < 0) //leave as is
			return;
		
		try{
			SignalFetcher amds = AugMDSFetcher.defaultInstance();		
			AUGSignal pwrSig = (AUGSignal)amds.getSig(new AUGSignalDesc("/aug/" + pulse + "/" + powerSignal));
			double pwrTime[] = (double[])pwrSig.getTVectorAsType(double.class);
			double pwrAll[][][] = (double[][][])pwrSig.getDataAsType(double.class);
			
			int idx = OneLiners.getNearestIndex(pwrTime, time);
			
			List<int[]> selectedBeams = new ArrayList<int[]>(8);
			
			int nBoxes = pwrAll.length;
			for(int i=0; i < nBoxes; i++){
				int nBeams = pwrAll[i].length;
				for(int j=0; j < nBeams; j++){
					if(pwrAll[i][j][idx] > beamOnPowerThreshold){
						selectedBeams.add(new int[]{i , j, autoSelectFuel });
					}
					
				}
			}
			
			this.selectedBeams = selectedBeams.toArray(new int[0][]);
			
		}catch(RuntimeException err){
			System.err.println("WARNING: No power signals for NBI, using existing selection:");
			err.printStackTrace();
			return;
		}
		
	}
	
	public GaussianBeamDefinition[] getBeamDefinitions(){ update(); return beamDefs; }
	
	public void setSelectedBeams(int[][] selectedBeams) { 
		set("selectedBeams", selectedBeams);
	}

	public int[][] getSelectedBeams() { update(); return this.selectedBeams; }

	public void setComponentPowerFractions(double powerFractions[][][]){
		set("componentPowerFractions", powerFractions);
	}
	
	public void setComponentPowerFractions(int beamBox, int fuelType, double powerFractions[]){
		update();
		componentPowerFractions[beamBox][fuelType] = powerFractions;
		set("componentPowerFractions", componentPowerFractions);
	}
	
	public double[] getComponentPowerFractions(int beamBox, int fuelType) { return this.componentPowerFractions[beamBox][fuelType]; }
	
	public void setBeamVelocityAngularSigmaAll(double velocityAngularSigma) {
		set("velocityAngularVariance", velocityAngularSigma * velocityAngularSigma);
	}
	
	public void setVelocityAngularVariance(double velocityAngularVariance){
		set("velocityAngularVariance", velocityAngularVariance);
	}
	
	public void setEnergyFractionalStdDeviation(double energyFractionalStdDeviation){
		set("energyFractionalStdDeviation", energyFractionalStdDeviation);
	}
	
	public void setAutoSelectFuel(int beamFuel){ this.autoSelectFuel = beamFuel; setChanged("autoSelectFuel"); }

	@Override
	public void updateState() {		
		// Check beam auto select
		if(selectedBeams == null || (isSet("pulse") && isSet("time") &&
					( isChanged("autoSelectFuel") || isChanged("matchGeomToNov2013Faro") ||
					isChanged("pulse") || isChanged("time")) ) ){
			
			updateBeamSelection();
			beamDefs = null;
		}
		
		if(beamDefs == null || isChanged("componentPowerFractions")){
			calcBeams();
		}
	}
	
	
	@Override
	public double[][] generatePoints(double l0, double l1, double rMax, int nReq) { //Regular
		update();
		
		int j = 0; //beamIdx, err...
		
		double right[] = Algorithms.reNorm(Algorithms.cross(beamDefs[j].uVec, new double[]{0,0,1}));
		double up[] = Algorithms.cross(right, beamDefs[j].uVec);
		
		//calc number of point along axis
		int nL = (int)Math.pow(nReq * (l1-l0)*(l1-l0) / (Math.PI * rMax*rMax), 1.0/3.0);
		int nxy = (int)Math.sqrt(4 * nReq / nL / Math.PI);
		double dl = (l1-l0) / nL;
		double dxy = 2*rMax / nxy;
		double d = (dl + dxy) / 2;
		nL = (int)((l1-l0) / d);
		nxy = (int)(2.0 * rMax / d);
		
		//count actual number in CSA
		int nCS = 0;
		for(int iX=0; iX < nxy; iX++){
			double x = -rMax + d * iX;
			for(int iY=0; iY < nxy; iY++){
				double y = -rMax + d * iY;
				if( x*x + y*y <= rMax*rMax )
					nCS++;
			}
		}
		
		int n = nL * nCS;
		double ret[][] = new double[3][n]; 
		
		int i=0;
		for(int iL=0; iL < nL; iL++) {
			double l = l0 + iL * d;
			for(int iX=0; iX < nxy; iX++){
				double x = -rMax + d * iX;
				for(int iY=0; iY < nxy; iY++){
					double y = -rMax + d * iY;
					if( x*x + y*y <= rMax*rMax ){
						ret[0][i] = beamDefs[j].focus[0] + l * beamDefs[j].uVec[0] + x * right[0] + y * up[0];
						ret[1][i] = beamDefs[j].focus[1] + l * beamDefs[j].uVec[1] + x * right[1] + y * up[1];
						ret[2][i] = beamDefs[j].focus[2] + l * beamDefs[j].uVec[2] + x * right[2] + y * up[2];

						i++;
					}
				}
			}
		}
		
		return ret;
	}
	
	
	public double[][] generatePointsRandom(double l0, double l1, double rMax, int n) {
		update();
		
		double ret[][] = new double[3][n]; 
		for(int i=0; i < n; i++){
			int j = (int)(RandomManager.instance().nextUniform(0, 1) * beamDefs.length);
			double l = l0 + RandomManager.instance().nextUniform(0, 1) * (l1 - l0);
			ret[0][i] = beamDefs[j].focus[0] + l * beamDefs[j].uVec[0];
			ret[1][i] = beamDefs[j].focus[1] + l * beamDefs[j].uVec[1];
			ret[2][i] = beamDefs[j].focus[2] + l * beamDefs[j].uVec[2];
			
			double r = RandomManager.instance().nextUniform(0, 1);
			r = FastMath.sqrt(r);
			double phi = RandomManager.instance().nextUniform(0, 1) * 2 * Math.PI;
			
			double right[] = Algorithms.reNorm(Algorithms.cross(beamDefs[j].uVec, new double[]{0,0,1}));
			double up[] = Algorithms.cross(right, beamDefs[j].uVec);

			ret[0][i] += r * Math.cos(phi) * up[0] + r * Math.sin(phi) * right[0];
			ret[1][i] += r * Math.cos(phi) * up[1] + r * Math.sin(phi) * right[1];
			ret[2][i] += r * Math.cos(phi) * up[2] + r * Math.sin(phi) * right[2];
			
		}
		return ret;
	}

	/** Output the beam focal point and unit vectors, for copying into the optics package */
	public static void main(String[] args) {

		AugHeatingBeams beamDefs = new AugHeatingBeams();		
		GaussianBeamDefinition beamDef[] = beamDefs.getBeamDefinitions();
	
		DecimalFormat fmt = new DecimalFormat("0.0000000");
		
		System.out.println("public static final double nbiStartAll[][] = { ");
		for(int i=0; i < beamDef.length; i++){
			System.out.println("\t{ " + fmt.format(beamDef[i].focus[0]) + 
								", " + fmt.format(beamDef[i].focus[1]) + 
								", " + fmt.format(beamDef[i].focus[2])
								+ " },");
			
		}
		
		System.out.println("};\n\npublic static final double nbiUnitAll[][] = {");
		for(int i=0; i < beamDef.length; i++){
			System.out.println("\t{ " + fmt.format(beamDef[i].uVec[0]) + 
								", " + fmt.format(beamDef[i].uVec[1]) + 
								", " + fmt.format(beamDef[i].uVec[2])
								+ " },");
		}
		System.out.println("};\n");
	}
}
